/**
*
* $Id: VELCRO.cpp,v 0.1 2004/11/20 18:03:45 erv Exp $
* (c) 2004 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* Autopilot by rjcroy
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/

#include "Velcro.h"



// 1. vertical lift component (wings and body)
// Borrowed from Mike Norman's Shuttle code

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*15.0;
	static const double istep = 1.0/step;
	static const int nabsc = 25;
	//                              -180 -165  -150 -135  -120 -105 -90 -75.0  -60.0  -45.0  -30.0  -15.0   0.0   15.0      30.0     45.0  60.0  75.0  90.0 105.0  120.0  135.0  150.0  165.0 180.0  
	static const double CL[nabsc] = {0.1, 0.17, 0.2, 0.2, 0.17, 0.1, 0, -0.11, -0.24, -0.38,  -0.5,  -0.5,  0.02, 0.6955,    0.66,   0.42, 0.24, 0.10, 0.0, -0.16, -0.26, -0.29, -0.24, -0.1, 0.1};
	static const double CM[nabsc] = {  0,    0,   0,   0,    0,   0, 0,     0,     0,  0.002,  0.004, 0.0025,0.0012,      0,-0.0012,-0.0007,    0,    0,   0,     0,     0,     0,     0,    0,   0};
	// lift and moment coefficients from -180 to 180 in 15 degree steps.
	// This uses a documented lift slope of 0.0437/deg, everything else is rather ad-hoc

	// change for drag increase at high Mach:
	// Subsonic Equivalents

	double Cl0, Cl1, Cd0, Cd1, Cd2;
	if(M>=5.0) {
		Cl0 = -0.187040;
		Cl1 = 1.675600;
		Cd0 = 0.068540;
		Cd1 = -0.21767;
		Cd2 = 1.3632;
		*cl = Cl0 + Cl1*aoa;
		*cd = Cd0 + Cd1*aoa + Cd2*aoa* aoa;
	}
 	aoa += PI;
	int idx = max (0, min (23, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cm = (CM[idx] + (CM[idx+1]-CM[idx])*d);
	if(M<1.25) { 
 		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
 		*cd = 0.04 + oapiGetInducedDrag (*cl, 2.1, 0.82);
	}
	else if(M<5.0) { 
 		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
 		*cd = 0.07 + oapiGetInducedDrag (*cl, 2.1, 0.65);
// 		*cd = 0.06 + oapiGetInducedDrag (*cl, 2.266, 0.6);
	}
}

/*
void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, 0*RAD, 15*RAD,25*RAD,30*RAD,60*RAD,180*RAD};
	static const double CL[nabsc]  = {       0,      0,   -0.4,     0,    0.7,     1,   0.8,     0,      0};
	static const double CM[nabsc]  = {       0,      0,  0.014, 0.004,-0.0074,-0.094,-0.012,     0,      0};
	for (int i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.015 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
	if (M > 1.0) *cd *= 2.2;
	else *cd *= 0.6;
}
*/

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 8;
	int i;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	
	if (M > 1.0) *cd *= 1.9;
	else *cd *= 0.6;
}

// 2. vertical lift component (horizontal stabilisers and body)

void VLiftCoeff2 (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 8;
	int i;
	static const double AOA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	
	if (M > 1.0) *cd *= 1.9;
	else *cd *= 0.6;
}

																					
void VELCRO::InitialiseGuidance()
{
	

//	time=0.0;

	pitchinput = 0.0;
	rollinput = 0.0;
	engineignited = false;
	last_simt=0.0;
	ignition_time=0.0;
	first_guidance_pass = true;
	Flydown = false;

	// rjcroy Guidance variables

	INCLINATION = 51.63 ;		// degrees

	guidance = false;			// guidance switch
	Alt_Flag = false;			// True if max launch altitude reached
	MECO_Cond = false;			// True if MECO conditions met
	First_Guidance_Pass = false;
	T_minus_10 = true;			// True at T minus 10 seconds.
	TZero = -100;				// Time Zero - Launch time
	PC_D = 0.9;					// Pitch control coefficient D for d_theta /d_t
	PC_K = 0.2;					//    "    "         "       K  "  theta
	YC_D = 0.5	;				// Yaw Control
	YC_K = 0.3;
	PC_D_FINAL = 0.9;			// Pitch control coefficients for final section.
	PC_K_FINAL = 0.6; 
	D_FINAL  = 0.8;				// Vertical velocity control coefficients
	K_FINAL = 0.8;
	ENG_THRST_2 = 0 ;		// Thrust rating of Stage 2 engine. (N)  
	i_last = 0;					// value of incliniation in previous iteration.
//	Planet_Name = new char[6];
//	strcpy( Planet_Name, "Earth");
	R_EARTH = 6371000;			// radius of the earth (m)
	U_EARTH = 3.986e5 * 1000 * 1000 * 1000; //( m^3/sec^2)
	Ob_EARTH = 23.45 * RAD ;	// Obliquity of the Earth (rad)
	APO_ALT = 232 ;				// Altitude of apoapsis
//	APO_ALT = 1705;

	//  inclination correction factor constant. (3580)
	ICF_K = 62.008 * INCLINATION*INCLINATION 
			- 6848.7*INCLINATION + 189807;	

	DisplayIncSP = false;		// toggle for displaying the inclination setpoint.

	NewIncSP = false ;			// True if a user has entered a new inclination SP
	PostMECO = false ;			// switch for cleaning up after MECO

	BadHeading = false;			// true if soyuz is at an azimuth not suitable for the G&CS



}

double VELCRO::GetCPitch(double t)
{
	int i;
	if (t>met[N-1]) return cpitch[N-1];
	i=1;
	while (met[i]<t) i++;
	return cpitch[i-1]+(cpitch[i]-cpitch[i-1])/(met[i]-met[i-1])*(t-met[i-1]);
}

// rjcroy Lateral Guidance System ( called by Guidance_Routine() ).

void VELCRO::Lateral_Guidance(VESSEL *vessel, double SimT, double SimT_Last)
{
	// Lateral Guidance System 

		double hdg, diff, rate, amount, targetrate, TgtHdg;

		if (GetAltitude() > 50.0) {
			targetrate = ((GetBank()*180/PI)-180)/-1.0;
			if (GetBank()<0) targetrate = -1.0*targetrate;

			if (energy > 0.20) {

				oapiGetHeading(GetHandle(),&hdg);
//				hdg = vst.arot.y * 180/PI;
				hdg *= 180/PI;
//				hdg += 180;

				GetRelativeVel (hEarth, Rel_Vel);
				GetEquPos( Longitude, Latitude, Radius );
				Velocity = sqrt((Rel_Vel.x*Rel_Vel.x) + (Rel_Vel.y*Rel_Vel.y) + (Rel_Vel.z*Rel_Vel.z));

				oapiGetRelativeVel(hVessel, hEarth, &Rel_Vel);
				oapiGetRelativePos(hVessel, hEarth, &Rel_Pos);
				
				// Change ecliptic rel. vectors to equatorial vectors.
 				temp.x = Rel_Vel.x;
				temp.y = Rel_Vel.z;
				temp.z = Rel_Vel.y;
				veccpy( Rel_Vel, temp);
				temp.x = Rel_Pos.x;
				temp.y = Rel_Pos.z;
				temp.z = Rel_Pos.y;
				veccpy( Rel_Pos, temp);
				temp.x = Rel_Pos.x ;
				temp.y = cos(-Ob_EARTH)* Rel_Pos.y + sin(-Ob_EARTH)*Rel_Pos.z;
				temp.z =  -sin(-Ob_EARTH)* Rel_Pos.y + cos(-Ob_EARTH)*Rel_Pos.z;
				veccpy( Rel_Pos, temp);
				temp.x = Rel_Vel.x ;
				temp.y = cos(-Ob_EARTH)* Rel_Vel.y + sin(-Ob_EARTH)*Rel_Vel.z;
				temp.z =  -sin(-Ob_EARTH)* Rel_Vel.y + cos(-Ob_EARTH)*Rel_Vel.z;
				veccpy( Rel_Vel, temp);

				h_vector = crossp(Rel_Vel, Rel_Pos);
				h_abs = sqrt((h_vector.x*h_vector.x) + (h_vector.y*h_vector.y) + (h_vector.z*h_vector.z));
				i = 180 - acos(h_vector.z / h_abs) * DEG;
				i_dot = (i - i_last) / (MET - SimT_Last) ;
				i_last = i;
				ICF = Velocity / ICF_K;

				// Calculated error in inclination as output signal
				i_err = (i - INCLINATION) * (150 * ICF) + 250 * ICF * i_dot;
				if (i_err > ICF )
				{	
					i_err = ICF;
				}
				else if ( i_err < -ICF) 
				{
					i_err = -ICF;
				}
							
/*				double Phrad;
					//= (PadHeading-180)*PI/180;

				if ((Rel_Vel.z < 110.0) || (bgoingsouth)) {
					Phrad = PI/2 + (PI/2 - asin(cos(INCLINATION*PI/180)/cos(Latitude)));
					if ((PadHeading-180) > 180)
						Phrad = 3*PI/2 + (3*PI/2 - asin(cos(INCLINATION*PI/180)/cos(Latitude)));
					TgtHdg = Phrad*DEG;
					TgtHdg -= i_err;
					bgoingsouth = true;
				}
				else {
					Phrad =  asin(cos(INCLINATION*PI/180)/cos(Latitude));
					TgtHdg = Phrad*DEG;
					TgtHdg += i_err;
				}

				if ((TgtHdg > 90.0) || (PadHeading-180 >= 90.0))
					bgoingsouth = true;

				if ((Latitude*DEG >= i*0.98) && (fabs(i-INCLINATION)<0.02))
					bgoingsouth = true;
*/
				TgtHdg = PadHeading - 180.0;


//				sprintf(oapiDebugString(),"Latitude %.2f  inc %.2f TGTINC %.2f going %1d",Latitude*DEG, i, INCLINATION, bgoingsouth);
//				sprintf(oapiDebugString(),"TgtHdg %.2f  inc %.2f TGTINC %.2f Phrad %.2f PadHdg %.2f tmpy %.2f",TgtHdg, i, INCLINATION, Phrad*DEG,PadHeading-180,Rel_Vel.z);

				diff = hdg - TgtHdg;
				if (RollAtt == 1) diff = hdg - TgtHdg + 180;

				Yaw_Error = diff;

//				diff = hdg - (PadHeading - 180);
				if( diff > 180.0 )
					diff -= 360.0;
				else if( diff < -180.0 )
					diff += 360.0;
				targetrate = diff/+1.0;
			}

			if( targetrate > 8.0 )
				targetrate = 8.0;
			else if( targetrate < -8.0 )
				targetrate = -8.0;


			rate = vst.vrot.y * 180/PI;
			amount = fabs (targetrate/8.0);
			if (energy > 0) amount = amount / (energy*20);

//		sprintf(oapiDebugString(),"Bank %.1f Heading %.1f diff %.1f Targetrate %.1f",GetBank()*180/PI,hdg,diff, targetrate);

			if( fabs( rate-targetrate ) > 0.05 ) {
				if( rate < targetrate ) {
//					IncThrusterGroupLevel( THGROUP_ATT_YAWRIGHT, 0.05 );
//					SetThrusterGroupLevel( THGROUP_ATT_YAWLEFT, 0.0 );
					yawinput = -0.5*amount;
				}
				else {
//					IncThrusterGroupLevel( THGROUP_ATT_YAWLEFT, 0.05);
//					SetThrusterGroupLevel( THGROUP_ATT_YAWRIGHT, 0.0 );
					yawinput = 0.5*amount;
				}
			}
			else {
//				SetThrusterGroupLevel( THGROUP_ATT_YAWLEFT, 0.0 );
//				SetThrusterGroupLevel( THGROUP_ATT_YAWRIGHT, 0.0 );
				yawinput = 0.0;
			}
		}
}

// rjcroy Guidance & Control System
void VELCRO::Guidance_Routine(VESSEL *vessel, double SimT, double SimT_Last)
{

	hVessel = GetHandle();
	hEarth = GetSurfaceRef ();
//	hEarth = oapiGetGbodyByName ( Planet_Name );


	if (T_minus_10)
	{
		// Record launch time
		if (TZero == 0){
			TZero = SimT + 0.25;
		}

		if (MET > -0.2 && MET < -0.1) {
//			oapiSetEngineLevel(hVessel, ENGINE_MAIN, 0.3);
//			LaunchClamps();
		}
		else if (MET > -0.1 && MET < 0) {
//			oapiSetEngineLevel(hVessel, ENGINE_MAIN, (0.2*MET + 1));
//			LaunchClamps();
			INCLINATION = 180 - acos(sin(PadHeading*PI/180)*cos(Latitude))*DEG;
		}
		else if (MET > 0)
		{
			T_minus_10 = false;
			First_Guidance_Pass = true;
			if (guidance)
				oapiSetEngineLevel(hVessel, ENGINE_MAIN, 1);

			GetStatus(vst);
	
			initialgamma = vst.arot.z * 180/PI;
			launchrotate =	PadHeading - initialheading;

			if( launchrotate < -180.0 )
				launchrotate += 360.0;
			else if( launchrotate > 180.0 )
				launchrotate -= 360.0;

		}
//		if (oapiGetFocusObject() == GetHandle())
//			sprintf(oapiDebugString(),"T - %.1f", fabs(MET));
	}		

	else if (First_Guidance_Pass) 
	{
		
		// Start the main engines
		if (guidance) {
			oapiSetEngineLevel(hVessel, ENGINE_MAIN, 1);
//			sprintf(oapiDebugString()," ");
		}

		// Unset the first pass flag
		First_Guidance_Pass = false;
	}

	else if (!Alt_Flag) 
	{
		if (guidance)
			oapiSetEngineLevel(hVessel, ENGINE_MAIN, 1);

		if (initialgamma == 0.0)
			initialgamma = vst.arot.z * 180/PI;

		// Get current velocity
		oapiGetRelativeVel(hVessel, hEarth, &Rel_Vel);
		Velocity = sqrt((Rel_Vel.x*Rel_Vel.x) + (Rel_Vel.y*Rel_Vel.y) + (Rel_Vel.z*Rel_Vel.z));

		// Get correct pitch from table
		energy = GetAltitude() + Velocity*Velocity;
		energy *= 1e-6;
		energy *= PitchMult;
		Pitch_SP=GetCPitch(energy);
		MJD_Time = oapiGetSimMJD();
		GetElements( elements, MJD_Time );
		Alt_Apoapsis = (elements.a * ( 1 + elements.e )/1000) - 6371; 
/*		if (Alt_Apoapsis > 170.0) {
			Pitch_SP = Pitch_SP/4.0;			// Flydown -- we've met our altitude goal
			if ((GetAOA()*DEG < 2.0))
				Pitch_SP += 1-1.0*GetAOA()*DEG;
		}
*/
		if (Alt_Apoapsis < 1.5*ScreenHeight)
			if (GetThrusterCount() < 22)			// Special low-power routine
				if (EngineThrust/9.81 < 0.6*GetMass()) {
					Pitch_SP += 8.0*((0.6*GetMass()) / (EngineThrust/9.81));
					if (Pitch_SP > 40.0) Pitch_SP = 40.0;
				}
		GetHorizonAirspeedVector(horiz);
		if ((Alt_Apoapsis > ScreenHeight) || (horiz.y < -10.0) || (Flydown)) {
			Flydown = true;

//			Pitch_SP *= 0.25;					// Flydown -- we've met our altitude goal
			double Pitch_zero = -0.012*horiz.y;
			double Aggressive = 0.98-(sqrt(energy)*0.0156);
//			double Aggressive = 0.85+(energy*0.002);
			if (Aggressive > 0.98) Aggressive = 0.98;
			if (horiz.y > 0.0) {
				Aggressive = 0.98-(horiz.y/1000);
				if (Aggressive < 0.85)
					Aggressive = 0.85;
			}
			if (RollAtt == 0)
				Pitch_SP += Pitch_zero-Aggressive*GetAOA()*DEG;	
			else if (RollAtt == 1)
				Pitch_SP += Pitch_zero+Aggressive*GetAOA()*DEG;
			if (Alt_Apoapsis < 1.5*ScreenHeight)
				if (GetThrusterCount() < 22)						// Special low-power addition
					if (EngineThrust/9.81 < 0.6*GetMass()) {
							double multiple = 8.0;
							multiple /= (horiz.y/25);
							Pitch_SP += multiple*((0.6*GetMass()) / (EngineThrust/9.81));
							if (Pitch_SP > 40.0) Pitch_SP = 40.0;
						}
			if (Pitch_SP > 40.0) Pitch_SP = 40.0;
		}
		else Pitch_SP += -0.02*horiz.y;
//		else if ((Alt_Apoapsis > 120.0) && (GetAOA()*DEG < 2.0))
//			Pitch_SP += 1-1.0*GetAOA()*DEG;


/*		if (Alt_Apoapsis < ScreenHeight)
			if (Pitch_SP < (ScreenHeight - Alt_Apoapsis)/5) {
				Flydown = true;
				Pitch_SP = (ScreenHeight - Alt_Apoapsis)/5;
			}
*/
		if (Pitch_SP < MinPitch) {
			Pitch_SP = MinPitch;
			Flydown = true;
		}

//		sprintf(oapiDebugString(),"Alt_Apoapsis %.1f Pitch_SP %.1f CPitch %.1f",Alt_Apoapsis, Pitch_SP, GetCPitch(MET));
//		sprintf(oapiDebugString(),"Energy %.3f Pitch_SP %.1f altitude %.1f airspeed %.1f",energy, Pitch_SP,GetAltitude(),Velocity);

		// Control to pitch setpoint
		Pitch=GetPitch(  );
		Pitch = Pitch * DEG;
		Pitch_Error = Pitch - Pitch_SP;
//		if (ROLL < 90.0)
//			Pitch_Error = Pitch_SP - Pitch;
		if (RollAtt == 1)
			Pitch_Error *= -1;

//		Pitch_Error *= -1.0*cos(GetBank());

		GetAngularVel( Angular_Velocity ); 
		Pitch_Error_Dot = Angular_Velocity.x * DEG; // get pitch change rate in degrees/s


		Pitch_Cntrl_Sig =  - PC_D * Pitch_Error_Dot + PC_K * Pitch_Error;
//		if ((MET > 10.0) &&(GetBank() < PI/2) && (GetBank() > -PI/2))
//			Pitch_Cntrl_Sig *= -1.0;

		Pitch_Error *= PC_K;

		if (energy < 0.19)
			Pitch_Cntrl_Sig = 0.0;

		if (Pitch_Cntrl_Sig > 1 )
		{
//			SetAttitudeRotLevel( 0, 1); // 0 is y-axis
			pitchinput =-1.0;
		}
		else if (Pitch_Cntrl_Sig < -1) 
		{
//			SetAttitudeRotLevel( 0, -1);
			pitchinput = +1.0;
		}
		else {
//			SetAttitudeRotLevel( 0 , Pitch_Cntrl_Sig);
			pitchinput = -Pitch_Cntrl_Sig;
		}
		

//		if (GetBank() > 0.0)
//			pitchinput = 0.0;

//		if ((GetBank() > 3*PI/2) || (GetBank() < -3*PI/2) || ((GetBank() < PI/2) && (GetBank() > -1*PI/2)))
//			pitchinput *= -1.0;

		Yaw_Error = 0.0;

		if (energy > 3.5) {
			diff = 1.0*GetBank()*180/PI;
			yawangle = DEG*GetSlipAngle();
			yawinput = yawangle / 15.0;
			if (oapiGetTimeAcceleration() > 10.0)
				yawinput *= 0.2;
		}
		else {
		// Heading control
			if (energy > 0.18)
				Lateral_Guidance(this, SimT, SimT_Last);
		}

		// Roll control
		// Adapted from Hendo's Ariane V


		if(energy > 0.20) {
				
//		vst.version = 2;
//		vst.flag = 0;
		vessel->GetStatus(vst);


		double target = initialgamma - launchrotate;

		if( target > 180.0 )
			target -= 360.0;
		else if( target < -180.0 )
			target += 360.0;
					
		double diff = vst.arot.z*180/PI - target;
		if( diff > 180.0 )
			diff -= 360.0;
		else if( diff < -180.0 )
			diff += 360.0;

//		if (energy > 3.0) diff = 0.0;
//		if ((MET > 150.0) && (fabs(fabs(GetBank()) - PI) < .2)) diff = GetBank()*180/PI + target;
//		if ((GetBank()*180/PI < 135) || (GetBank()*180/PI > 225)) {
//		if ((ConstBank) || ((energy > 0.3) && (fabs(fabs(GetBank()) - PI) < .05))) {
//			ConstBank = true;
//			if (oapiGetFocusObject() == GetHandle()) sprintf(oapiDebugString(),"Bank %.1f ",GetBank()*DEG);
			if (GetBank() > 0.0)
				diff = -(GetBank()-PI)*10;
			else diff = -(GetBank()+PI)*10;
//			if (ROLL < 90.0) diff = -1.0*GetBank()*180/PI;
//		}
//			diff *= -1.0;
//		}

//		if (energy > 1.0) diff = GetBank()*180/PI + target;
//		if (MET > 250) diff = 0.0;


//		sprintf(oapiDebugString(),"initialhdg %.1f initialgamma %.1f launchrotate %.1f target %.1f diff %.1f TGT_ %.1f",initialheading,initialgamma, launchrotate, target,diff, PadHeading);


		if (RollAtt == 1) {
//			yawinput *= -1;
			diff = -GetBank() * 10;
		}
		double targetrate = diff/-1.0;
		if( targetrate > 8.0 )
			targetrate = 8.0;
		else if( targetrate < -8.0 )
			targetrate = -8.0;

		if (GetBank() > 0.0)
			Bank_Error = (GetBank()-PI);
		else Bank_Error = (GetBank()+PI);
		if (RollAtt == 1)
			Bank_Error = GetBank();


		double rate = vst.vrot.z * 180/PI;
		double amount = fabs (targetrate/4.0);
//		if (energy > 0.25) 
//			amount = amount / (4*energy);

//		if (MET > 0) amount = amount / (MET/10);

		if( fabs( rate-targetrate ) > 0.05 ) {
			if( rate < targetrate ) {
//				IncThrusterGroupLevel( THGROUP_ATT_BANKRIGHT, 0.05 );
//				SetThrusterGroupLevel( THGROUP_ATT_BANKLEFT, 0.0 );
				rollinput = -1*amount;
			}
			else {
//				IncThrusterGroupLevel( THGROUP_ATT_BANKLEFT, 0.05);
//				SetThrusterGroupLevel( THGROUP_ATT_BANKRIGHT, 0.0 );
				rollinput = amount;
			}
		}
		else {
//			SetThrusterGroupLevel( THGROUP_ATT_BANKLEFT, 0.0 );
//			SetThrusterGroupLevel( THGROUP_ATT_BANKRIGHT, 0.0 );
			rollinput = 0.0;
		}

//		if ((MET > 60.0) && (diff < 0.5))

//		if (MET > 60.0)
//			Lateral_Guidance(this, SimT, SimT_Last);
//		else
//			yawinput = 0.0;

//		GetAngularVel(AngularVel);

//		yawinput=(-20*AngularVel.y);		

		// END Hendo's roll control


		}

		Altitude=GetAltitude();
		a = ((APO_ALT * 1000 + R_EARTH) + (Altitude + R_EARTH)) / 2;
		e = ((APO_ALT * 1000 + R_EARTH) / a ) - 1;
		h = sqrt(U_EARTH * a * ( 1 - (e * e)));
		V_periapsis = h / (Altitude + R_EARTH);

		
		// Get current velocity
		GetRelativeVel (hEarth, Rel_Vel);
		Velocity = sqrt((Rel_Vel.x*Rel_Vel.x) + (Rel_Vel.y*Rel_Vel.y) + (Rel_Vel.z*Rel_Vel.z));

		if ((energy > 10.0) && (Velocity > V_periapsis)) 
		{
			MECO_Cond = true;
			TimeMECO = SimT;
			oapiSetEngineLevel(hVessel, ENGINE_MAIN, 0);
			Guidance(false);
			pitchinput = 0.0;
			yawinput = 0.0;
			rollinput = 0.0;
		}

		
	}

	else if (MECO_Cond && PostMECO) 
	{
		if ((SimT - TimeMECO) < 10) {
//			sprintf(oapiDebugString(), "Guidance OFF.  Inclination (Equatorial): %.3f", i); 
		}	
		else {
			guidance = false;
//			sprintf(oapiDebugString(), "");
		}
	}
}
// END Guidance_Routine()


void VELCRO::GetInclinationSP()
{
	// Gets a new Inclination setting from the pilot
	bool AzInput (void *id, char *str, void *data);
	
//	oapiOpenInputBox ("Inclination Setpoint, e.g. 5163 = 51.63 deg (47 to 56)", IncInput, 0, 25);
	oapiOpenInputBox ("Enter launch azimuth, 0 to 90 (east) to 360", AzInput, 0, 20, (void*)this);

//	GetEquPos( Longitude, Latitude, Radius );
//	sprintf(oapiDebugString(), "Target inclination (Equatorial): %.3f", acos(sin(PadHeading*PI/180)*cos(Latitude))*DEG); 
}


bool AzInput (void *id, char *str, void *data)
{
//	sscanf(str, "%4lf", &PadHeading);
//	PadHeading = PadHeading / 100.0;
//	PadHeading += 180.0;
/*	sscanf(str, "%4lf", &INCLINATION);
	if (INCLINATION > 4700 && INCLINATION < 5600 ) {
		INCLINATION = INCLINATION / 100;
		NewIncSP = true;
	}
*/
	return ((VELCRO*)data)->IncInput (str);
}

bool VELCRO::IncInput (char *rstr)
{
	sscanf(rstr, "%5lf", &PadHeading);
//	PadHeading = PadHeading / 100.0;
	PadHeading += 180.0;
	d_heading = PadHeading;
	displaytime = 0.01;
	return true;
}

// END rjcroy Guidance System


VELCRO::VELCRO (OBJHANDLE hObj, int fmodel)
: COMMON_VESSEL_EX (hObj)
{
//	hS1C				= oapiLoadMeshGlobal ("S1C");

	tex = oapiRegisterExhaustTexture("Exhaust_atsme");
	texstor = oapiRegisterExhaustTexture("RealExhaust");
	texkero = oapiRegisterExhaustTexture ("KeroExhaust");
	texfluor = oapiRegisterExhaustTexture ("FluorineExhaust");

	/* set the propellant handles to NULL so I can check if I need to 
	/* create them in the SetMainStage() function. */
	ph_first=NULL; 
	ph_Velcro=NULL;
	ph_otherdef=NULL;
	ph_PassToTank=NULL;
	AtchV = NULL;
	DetV = NULL;
	Vold = _V(0,0,0);
	IgniteNext = true;
	EnginesDone = false;
	atchmass = 0.0;
	secmass = 0.0;
	numengines = 1;
	mode = 0;
	endlevel = 0.0;
	dt = 0.0;
	simt = 0.0;
	oldsimt = -0.01;
	displaytime = 0.01;
	SimT_Last = 0.0;
	PadHeading = 999.0;
	last_simt=0.0;
	ignition_time=0.0;
	initialgamma = 0.0;
	pitchinput = 0.0;
	yawinput = 0.0;
	rollinput = 0.0;
	stage = 0;
	numdocks = 0;
	chosenpoint = 0;
	chosenpoint2 = 4;
	tgtpos = _V(0,0,-10);
	tgtdir = _V(0,0,0);
	tgtrot = _V(0,0,0);
	hV = NULL;
	detV = NULL;
	switchfocus = 0;
	DetachMe = false;
	DetachMeDone = false;
	ReInit = false;
	JettisonCount = 0;
	DoEnginesLater = 0;
	OnGround = false;
	ForceNext = false;
	Mass = -1.0;
	OriginalFuelMass = 0.0;
	EngineThrust = 0.0;
	IspVac = 0.0;
	IspSL = -1.0;
	numexhausts = 0;
	PrimeBooster = true;
	attachpos[0] = NULL;
	attachpos[1] = NULL;
	attachpos[2] = NULL;
	attachpos[3] = NULL;
	NoJett = false;
	CAMERA_OFS = _V(-100,-100,-100);
	FairingNum = 0;
	FairingJettPress = 0.0;
	FairingJett = _V(0,0,0);
	Thrusting = false;
	DelayStart = -999.0;
	JettOnMECO = true;
	JettOnEmpty = 1;
	padbias = 0.0;
	CountingDown = false;
	CenterThrust = false;
	MEngineVec = _V(0,0,1);
	PassToTank = 0;
	FeedRate = 1.0;
	Atchflowrate = 0.0;
	ThrottleBack = 99999.9;
	RollAtt = 0;
	Bank_Error = 0.0;
	Yaw_Error = 0.0;
	LEStime = 0.0;
	NegativeSteer = false;
	Wing.Location = _V(0,0,0);
	Wing.span = -1.0;
	Wing.S = -1.0;
	Wing.AR = -1.0;
	HStab.Location = _V(0,0,0);
	HStab.span = -1.0;
	HStab.S = -1.0;
	HStab.AR = -1.0;
	VStab.Location = _V(0,0,0);
	VStab.span = -1.0;
	VStab.S = -1.0;
	VStab.AR = -1.0;
	ScreenHeight = 125.0;
	PitchMult = 1.0;
	MinPitch = 0.0;
	fuellevel = 0.0;
	IgniteWithCore = true;
	PayAtch = false;
	Tipoff = 1.0;
	SepForce = 1.0;
	GetDocksDone = false;
	GetDockingPorts = false;
	BecomeSomethingElse = false;
	mayfly = -999.0;
	Guid = false;
	ShiftParentCOM = _V(0,0,0);
	ShiftFocus = 1;

	//		     			   child       pos           dir       rot      handle
//	Baypos = CreateAttachment (false, _V(0,0,0.0),   _V(0,0,1), _V(0,-1,0), "X");


/*	int i;
	char classname[30];
	strncpy( classname, GetClassName(), 29 );
	classname[29]=0;
	for( i = 0 ; i < strlen( classname ) ; i++ ) {
		classname[i] = toupper(classname[i]);
	}
*/
/*	if( strstr( classname, "VELCRO" ) )
		stage = 0;
	else if( strstr( classname, "S1C" ) ) {
		S1C = true;
		stage = 0;
	}
	else
*/
	stage = 0;

/*
	attach[0] = _V(0,0,22.8);
	if (S1C) attach[0] = _V(0,0,18.4);
	attachdir[0] = _V(0,0,1);
	attachrot[0] = _V(0, 1,0);
	//	  		       			     child   pos           dir           rot      handle
	attachpos[0] = CreateAttachment (false, attach[0],   attachdir[0], attachrot[0], "X");

	attach[1] = _V(0,-7.0,-7.0);
	attachdir[1] = _V(0,0,1);
	attachrot[1] = _V(0, 1,0);
	//	  		       			     child   pos           dir           rot      handle
	attachpos[1] = CreateAttachment (false, attach[1],   attachdir[1], attachrot[1], "X");

	attach[2] = _V(0,0,22.8);
	if (S1C) attach[2] = _V(0,0,18.4);
	attachdir[2] = _V(0,0,1);
	attachrot[2] = _V(0, 1,0);
	//	  		       			     child   pos           dir           rot      handle
	attachpos[2] = CreateAttachment (true, attach[2],   attachdir[2], attachrot[2], "X");

	attach[3] = _V(0,-7.0,-7.0);
	attachdir[3] = _V(0,0,1);
	attachrot[3] = _V(0, 1,0);
	//	  		       			     child   pos           dir           rot      handle
	attachpos[3] = CreateAttachment (true, attach[3],   attachdir[3], attachrot[3], "X");

*/

	MyID=ConnectToOrbiterSoundDLL3(GetHandle());
//		havesound=false;
//	else
//	{
		havesound=true;
		RequestLoadVesselWave3(MyID,PLAYGEAR,"Sound\\Velcro\\gear.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYBANG,"Sound\\Velcro\\stg2sep.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYJETT,"Sound\\Velcro\\fasep.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYJET,"Sound\\Velcro\\jet.wav",BOTHVIEW_FADED_FAR);
//	}

	CheckCentralMessage();   

	Firstsweep = true;
	Secondsweep = false;

}


VELCRO::~VELCRO()
{

}


void VELCRO::TimeStep(  double simt )
{
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GPDS requests and returns a bool (true/false)
	/* if any GPDS requests have been implemented.
	/* If any have (eg adding a new payload), you will need to re-draw your ship, hence the
	/* call to the (repeatable) SetMainStage(); */
	if( CheckCentralMessage() ) {
		while( CheckCentralMessage() ) ;
			SetMainStage();
	}

//	if (oapiGetTimeAcceleration() > 100.0) return;

	dt = simt - oldsimt;
	oldsimt = simt;
	if ((displaytime > 0) && (oapiGetFocusObject() == GetHandle())) displaytime += dt;


	if ((stage > 0) && (mayfly > -999.0)) {
		mayfly -= dt;
		if (mayfly < 0.0)
			oapiDeleteVessel(GetHandle());
	}

	if ((stage == 0) && (mode == 4) && (EnginesStarted) && (EnginesDone)) {				// if CROSSFEED,
																						// transfer fuel across!
		if (ph_PassToTank == NULL)
			ph_PassToTank=AtchV->GetPropellantHandleByIndex(PassToTank);

		Atchflowrate = AtchV->GetPropellantFlowrate(ph_PassToTank); 
		Atchflowrate *= dt;
//		Atchflowrate = AtchV->GetTotalPropellantFlowrate() - AtchV->GetPropellantFlowrate(ph_first);  // Old method; fed strap-ons
		if (Atchflowrate <= 0.0)																					// Failed somehow (subsequent tank?)
			Atchflowrate = AtchV->GetPropellantMaxMass(ph_PassToTank) - AtchV->GetPropellantMass(ph_PassToTank);	// Use backup method!
		Atchflowrate *= FeedRate;
		if (Atchflowrate > AtchV->GetPropellantMass(ph_first))
			Atchflowrate = AtchV->GetPropellantMass(ph_first);

	}


	if (JettisonCount < 0) JettisonCount++;
	if (JettisonCount == -1) ReInit = true;

	if ((!guidance) && (displaytime < 5) && (GetAltitude() < 100.0) && (stage < 1) && (oapiGetFocusObject() == GetHandle()))
	{
		GetEquPos( Longitude, Latitude, Radius );
		sprintf(oapiDebugString(), "Target inclination (Equatorial): %.3f", (180 - acos(sin(d_heading*PI/180)*cos(Latitude))*DEG));

	}
	if (!guidance && !MECO_Cond ) {
		if(!NewIncSP) {
			// Set the Inclination setpoint for the guidance computer
			GetEquPos( Longitude, Latitude, Radius );
				INCLINATION = acos(Clairaut_Constant)*DEG;
		}
	}	

	if ((!Firstsweep) && (!Secondsweep)) {
	//Check if guidance is on and if so, call
	if (guidance)
	{
		{
			// Calculate MET
			if (TZero == -100) 
			{
				MET = -5 ;
				TZero = 0;
			}
			else 
			{
				MET = simt - TZero;
			}
			// Call the guidance routine
			Guidance_Routine(this, simt, SimT_Last);
			SimT_Last = MET;
		}
	}
	else if (!MECO_Cond && !PostMECO && (stage < 1) && (GetAltitude() > GetSize())) {
			if (TZero == -100) 
			{
				MET = -5 ;
				TZero = 0;
			}
			else 
			{
				MET = simt - TZero;
			}
			Guidance_Routine(this, simt, SimT_Last);
	}
	}

	if (mode < 2) {
		if (((GetThrusterMax0(th_first[0])*numengines)/GetMass()) > ThrottleBack) 
			for (int i = 0; i < numengines; i++) 
				SetThrusterMax0(th_first[i],0.98*GetThrusterMax0(th_first[i]));
	}

	if (ForceNext) {													// Post-jettison, we want to addforce to push us away
		ForceNext = false;												// but it has to be after everyone is separated
		if (mode > 0) {
			if (chosenpoint == 0) 
				AddForce (_V(0, 0,-100*SepForce*GetMass()), _V(0, 0, 0));
//				AddForce (_V(0, 0,-100*GetMass()), _V(0, 0, 0));
			else {
				VECTOR3 pos, dir, rot;
				GetAttachmentParams (attachpos[1], pos, dir, rot);
				AddForce (_V(200*SepForce*GetMass()*rot.x, 200*SepForce*GetMass()*rot.y, 0), _V(0, 0, Tipoff*GetSize()/20));
			}
		}
	}

	if (DoEnginesLater > 0)
		DoEnginesLater++;
	if (DoEnginesLater > 3)
		DoEngines();

	VectorThrust();

	if (ReInit) {
		double throttlesetting = GetThrusterGroupLevel(THGROUP_MAIN);

		if (AtchV != NULL)
			if ((ShiftParentCOM.x != 0.0) || (ShiftParentCOM.y != 0.0) || (ShiftParentCOM.z != 0.0)) {
				AtchV->ShiftCG(_V(-1.0*ShiftParentCOM.x,-1.0*ShiftParentCOM.y,-1.0*ShiftParentCOM.z));
				ShiftParentCOM = _V(0,0,0);
			}

		ReInit = false;
		ForceNext = true;
		SetMainStage();
		SetThrusterGroupLevel(THGROUP_MAIN,throttlesetting);
/*		if (AtchV != NULL)
		if ((ShiftParentCOM.x != 0.0) || (ShiftParentCOM.y != 0.0) || (ShiftParentCOM.z != 0.0)) {
//			AtchV->ShiftCG(_V(-1.0*ShiftParentCOM.x,-1.0*ShiftParentCOM.y,-1.0*ShiftParentCOM.z));
			VESSELSTATUS2 vs;
			AtchV->GetStatusEx (&vs);
			vs.vrot.x = 0.0;
			vs.vrot.y = 0.0;
			vs.vrot.z = 0.0;
			AtchV->DefSetStateEx (&vs);
		}
*/	}

	if (mode < 2) {
		if (stage == 0)
			if (GetPropellantMass(ph_first) == 0) 
				if (JettOnEmpty == 1) {
					Thrusting = true;
	//				JettisonNextStep = 1;
					JettisonPart();

					/* IMPORTANT */
					/* Always re-set the stage after any kind of change */
	//				SetMainStage();
				}
	} else {
		if ((stage == 0) && (DetachMeDone)) {
			if (AtchV->GetPropellantMass(ph_first) == 0) 
				if (JettOnEmpty == 1) {
					JettisonPart();
		
					/* IMPORTANT */
					/* Always re-set the stage after any kind of change */
	//				SetMainStage();
				}
		}
		else if ((DetachMeDone) && (switchfocus == 0) && (!IgniteWithCore))
			if (AtchV->GetPropellantMass(ph_first) == 0) {
				Thrusting = true;
				JettisonPart();
				Thrusting = true;
			}
	}


	if ((stage == 0) && (mode > 1))
		if ((EnginesStarted) && (AtchV->GetThrusterGroupLevel(THGROUP_MAIN) < 0.001) && (AtchV->GetThrusterGroupLevel(THGROUP_MAIN) != Smoldering)) {
			if (JettOnMECO) JettisonPart();
		}

/*	if ((stage == 0) && (mode < 2))
		if ((EnginesStarted) && (GetThrusterGroupLevel(THGROUP_MAIN) < 0.001) && (GetThrusterGroupLevel(THGROUP_MAIN) != Smoldering)) {
			if (JettOnMECO) JettisonPart();
		}
*/
	if ((stage == 0) && (mode == 3))													// if we're a booster,
		if ((EnginesStarted) && (EnginesDone))											// check to see if we've reached BECO fuel level
			if ((AtchV->GetPropellantMass(ph_otherdef) / AtchV->GetPropellantMaxMass(ph_otherdef)) < endlevel) {
				JettisonPart();
		}



	if (FairingNum > 0)
		if (GetAtmPressure() < FairingJettPress) {
			FairingJettison(FairingNum,FairingJett);
			FairingNum = 0;
		}

	

	if ((LEStime > 0.0) && (!Firstsweep) && (!Secondsweep)) {
		if (GetThrusterGroupLevel(THGROUP_MAIN) > 0.1)
			LEStime -= dt;
		if (LEStime < 0.0)
			JettisonLES();
	}

//	sprintf(oapiDebugString(),"DelayStart %.4f", DelayStart);

	if (GetThrusterGroupLevel(THGROUP_MAIN) > 0.0001)
		CountingDown = true;

	if (oapiGetTimeAcceleration() > 1000.0) {				// kludge to keep the throttle from
		SetThrusterGroupLevel(THGROUP_MAIN,0.0);			// running away at high time
		CountingDown = false;								// accelerations
	}

	if (mode < 2)
		if ((GetThrusterGroupLevel(THGROUP_MAIN) > (Smoldering/2)) || (CountingDown)) {
			if (DelayStart > 0.0) {
				SetThrusterGroupLevel(thg_first,Smoldering);
				DelayStart -= dt;
				CountingDown = true;
			} 
			else if ((DelayStart > -2*dt) || (ExhaustType == SRM)) {
				if (stage == 0) SetThrusterGroupLevel(thg_first,1.0);
				DelayStart = -999.0;
			}
		}


	if ((mode > 1) && (!Secondsweep) && (!Firstsweep)) {
		if ((IgniteWithCore) && (stage == 0)) {
			if (AtchV->GetThrusterGroupLevel(THGROUP_MAIN) > 0.001) {
				if (DelayStart > 0.0) {
					AtchV->SetThrusterGroupLevel(thg_first,0.0);
					DelayStart -= dt;
				}
				else
					AtchV->SetThrusterGroupLevel(thg_first,1.0);
				EnginesStarted = true;
				if (!DetachMeDone) DetachMe = true;
			}
		}
		else if (!IgniteWithCore)
			if ((EnginesStarted) || (GetThrusterGroupLevel(THGROUP_MAIN) > 0.001)) {
				EnginesStarted = true;
				if (ExhaustType == SRM)
					SetThrusterGroupLevel(thg_first,1.0);
			}
	}

	
/*	if ((mode > 1) && (!Secondsweep) && (!Firstsweep)) 
		if ((IgniteWithCore) && (stage == 0)) {
			if ((AtchV->GetThrusterGroupLevel(THGROUP_MAIN) > 0.001) || (AtchV->GetThrusterGroupLevel(THGROUP_MAIN) == Smoldering)) {
				if (DelayStart > 0.0) {
					AtchV->SetThrusterGroupLevel(thg_first,0.0);
					DelayStart -= dt;
				}
				else
					AtchV->SetThrusterGroupLevel(thg_first,1.0);				
			EnginesStarted = true;
			if (!DetachMeDone) DetachMe = true;
			}	
		}
		else {
			double thrustsetting = GetThrusterGroupLevel(THGROUP_MAIN);
			if ((EnginesStarted) && (ExhaustType == SRM)) thrustsetting = 1.0;
			AtchV->SetThrusterGroupLevel(thg_first,thrustsetting);
			if (thrustsetting > 0.001) {
				EnginesStarted = true;
				if (!DetachMeDone) DetachMe = true;				
			}
		}
*/



	if (switchfocus > 0) switchfocus ++;
	if (switchfocus > 3) {
		switchfocus = 0;
		if (mode == 0) {
			hV=oapiGetVesselByName(AtchV->GetName());
			if ((oapiGetVesselByName(GetName()) == oapiGetFocusObject()) && (ShiftFocus == 1))
				oapiSetFocusObject(hV);
			if ((Thrusting) && (IgniteNext))
				if (guidance)
					AtchV->SetThrusterGroupLevel(THGROUP_MAIN,GuidanceTransfer);
				else
 					AtchV->SetThrusterGroupLevel(THGROUP_MAIN,1.0);
			Thrusting = false;
		}
		else if (mode == 1) {
			hV=oapiGetVesselByName(AtchV->GetName());
			if ((oapiGetVesselByName(GetName()) == oapiGetFocusObject()) && (ShiftFocus == 1))
				oapiSetFocusObject(hV);
			if ((Thrusting) && (IgniteNext))
				if (guidance)
					AtchV->SetThrusterGroupLevel(THGROUP_MAIN,GuidanceTransfer);
				else
 					AtchV->SetThrusterGroupLevel(THGROUP_MAIN,1.0);
			Thrusting = false;
		}
		else if ((mode > 1) && (stage == 1) && (DetV != NULL) && (!IgniteWithCore)) {
			detV=oapiGetVesselByName(DetV->GetName());
			if ((oapiGetVesselByName(GetName()) == oapiGetFocusObject()) && (ShiftFocus == 1))
				oapiSetFocusObject(detV);
//			if ((Thrusting) && (IgniteNext))
				if (guidance)
					DetV->SetThrusterGroupLevel(THGROUP_MAIN,GuidanceTransfer);
				else
 					DetV->SetThrusterGroupLevel(THGROUP_MAIN,1.0);
			Thrusting = false;
//			AttachedStage = false;
			}

		Guidance( false );
		SetAttitudeRotLevel(_V(0,0,0));
		pitchinput = 0.0;
		yawinput = 0.0;
		rollinput = 0.0;
		if ((BecomeSomethingElse) && (strlen(becomename) > 1)) {
			char name[256];
			VESSELSTATUS vs;
			GetStatus (vs);
			OBJHANDLE hVess, rVess;
			VESSEL *rV;
			hVess = GetHandle();
			strcpy (name, GetName()); 
			rVess = oapiCreateVessel (name, becomename, vs);
			rV = oapiGetVesselInterface (rVess);
			if (oapiGetVesselByName(GetName()) == oapiGetFocusObject())
				oapiSetFocusObject(rVess);
			rV->SetFuelMass(GetPropellantMass(ph_first));

// Replicates the ForceNext function, which wouldn't be activated when BecomeSomethingElse deletes the vessel.
			if (mode > 0) {
				if (chosenpoint == 0) 
					rV->AddForce (_V(0, 0,-100*SepForce*GetMass()), _V(0, 0, 0));
//					AddForce (_V(0, 0,-100*GetMass()), _V(0, 0, 0));
				else {
					VECTOR3 pos, dir, rot;
					GetAttachmentParams (attachpos[1], pos, dir, rot);
					rV->AddForce (_V(200*SepForce*GetMass()*rot.x, 200*SepForce*GetMass()*rot.y, 0), _V(0, 0, Tipoff*GetSize()/20));
				}
			}

			oapiDeleteVessel(hVess);

		}
	}


	if (DetachMe) {
		DetachChild (attachpos[chosenpoint]);
//		if (OnGround)
//			AtchV->AddForce (_V(0, 0, 10*(AtchV->GetMass())), _V(0, 0, 0));
		tgtatch = AtchV->CreateAttachment (false, tgtpos,   tgtdir,        tgtrot, "BOOSTERP");
		MeV = oapiGetVesselByName(GetName());
		VESSELSTATUS vs;
		GetStatus(vs);
		vs.status = 0;
		DefSetState(&vs);
		AtchV->AttachChild (MeV, tgtatch, attachpos[chosenpoint+2]);
		SetAttachedStage();
		DetachMe = false;
		DetachMeDone = true;
		ForceMe = false;
	}

	if ((GetDockingPorts) && (!GetDocksDone))
		if ((!Firstsweep) && (!Secondsweep))
			if (GetAltitude() > 1000.0) {
				GetDocks();
				GetDocksDone = true;
			}

	if (ForceMe) {
//		if (OnGround) {
//			AddForce (_V(0, 0, 10*(GetMass())), _V(0, 0, 0));
//			AtchV->AddForce (_V(0, 0, 10*(AtchV->GetMass())), _V(0, 0, 0));
//		}
		DetachMe = true;
	}

	if (Secondsweep) {
		Secondsweep = false;

		if (PadHeading == 999.0) {
			oapiGetHeading(GetHandle(),&PadHeading);
			PadHeading *= 180.0/PI;
			PadHeading += 180.0;
		}

		if (Guid)
			Guidance(true);

		if ((energy > 10.0) && (Velocity > V_periapsis)) 
		{
			MECO_Cond = true;
			Guidance(false);
			pitchinput = 0.0;
			yawinput = 0.0;
			rollinput = 0.0;
		}

		for (i = 0 ; i < AttachmentCount (false); i++ ) {
			ATTACHMENTHANDLE hAtt = GetAttachmentHandle (false, i);
			const char *id = GetAttachmentId (hAtt);
			if (!strncmp (id, "VELCRO1", 7)) 
				attachpos[0] = GetAttachmentHandle (false, i);
			else if (!strncmp (id, "VELCRO2", 7)) 
				attachpos[1] = GetAttachmentHandle (false, i);
		}

		for (i = 0 ; i < AttachmentCount (true); i++ ) {
			ATTACHMENTHANDLE hAtt = GetAttachmentHandle (true, i);
			const char *id = GetAttachmentId (hAtt);
			if (!strncmp (id, "VELCRO3", 7)) 
				attachpos[2] = GetAttachmentHandle (true, i);	
			else if (!strncmp (id, "VELCRO4", 7)) 
				attachpos[3] = GetAttachmentHandle (true, i);
		}
	
			VESSELSTATUS vs;
			GetStatus(vs);
			if (vs.status == 1)
				OnGround = true;

			if (PayAtch) {				// Attach the secondary payload, if any.
				secV = oapiGetVesselByName(tgtname2);
				if (secV != 0) {
					SecV = oapiGetVesselInterface (secV);
					secmass = SecV->GetMass();	
					TgtSecPayload = SecV->CreateAttachment (true, tgtpos2,   tgtdir2,       tgtrot2, "X");
					if (SecV->GetAttachmentStatus(SecPayload) == NULL) 
						AttachChild (secV, SecPayload, TgtSecPayload);
					SetMainStage();
				}
			}
			if (mode > 0) {
			hV = oapiGetVesselByName(tgtname);
			if (hV != 0) {
				AtchV = oapiGetVesselInterface (hV);
				atchmass = AtchV->GetMass();	
				originalemptymass = AtchV->GetEmptyMass();	
				if (mode == 1) {
					//	  		       			       child   pos           dir           rot      handle
					tgtatch = AtchV->CreateAttachment (true, tgtpos,   tgtdir,       tgtrot, "BOOSTERS");

					if (AtchV->GetAttachmentStatus(attachpos[chosenpoint]) == NULL) 
						AttachChild (hV, attachpos[chosenpoint], tgtatch);
					SetMainStage();

				}
				if (mode > 1) {
					//	  		       			       child   pos           dir           rot      handle
/*					tgtatch = AtchV->CreateAttachment (false, tgtpos,   _V(0,0,-1),        tgtrot, "X");

					MeV = oapiGetVesselByName(GetName());
					AtchV->AttachChild (MeV, tgtatch, attachpos[chosenpoint+2]);
*/
					if (vs.status == 1) {
						vs.status = 0;
						DefSetState(&vs);
					}
					if (PrimeBooster == 0)
						ForceMe = true;
					else {
						DWORD nAttach = AtchV->AttachmentCount (true);
						for (DWORD j = 0; j < nAttach; j++) { // now scan all attachment points of the candidate
							ATTACHMENTHANDLE hAtt = AtchV->GetAttachmentHandle (true, j);
							const char *id = AtchV->GetAttachmentId (hAtt);
							if (!strncmp (id, "BOOSTERS", 8)) {			// We're the second (or later) booster -- somebody's already been here!
								ForceMe = true;							// Skip the attaching in Mode 1, and go straight to Mode 2 next step!
							}
						}
					}


					if (!ForceMe) {

						//	  		       			       child   pos           dir           rot      handle
						tempatch = AtchV->CreateAttachment (true, tgtpos,   tgtdir,       tgtrot, "BOOSTERS");
					if (AtchV->GetAttachmentStatus(attachpos[chosenpoint]) == NULL)
						AttachChild (hV, attachpos[chosenpoint], tempatch);
					}

					SetVelcroStage();

				}
			}
			if (mode > 0)
				if ((ShiftParentCOM.x != 0.0) || (ShiftParentCOM.y != 0.0) || (ShiftParentCOM.z != 0.0))
					AtchV->ShiftCG(ShiftParentCOM);

		}
	}

	if ((stage == 0) && (mode == 4) && (EnginesStarted) && (EnginesDone)) {				// if CROSSFEED,
																						// transfer fuel across!
		if ((AtchV->GetFuelMass()+Atchflowrate) <= AtchV->GetMaxFuelMass()) {
			AtchV->SetFuelMass(AtchV->GetFuelMass()+Atchflowrate);
			AtchV->SetPropellantMass(ph_first,GetPropellantMass(ph_first)-Atchflowrate);
		}
	}

/*
		if ((AtchV->GetFuelMass()+Atchflowrate) < AtchV->GetMaxFuelMass())
			AtchV->SetFuelMass(GetFuelMass()+Atchflowrate);
		AtchV->SetPropellantMass(ph_first,GetPropellantMass(ph_first)-Atchflowrate);
		AtchV->SetFuelMass(GetFuelMass());
		AtchV->SetPropellantMass(ph_first,GetPropellantMass(ph_first)-0.01);
*/

	if (GetThrusterGroupLevel(THGROUP_MAIN) == GuidanceTransfer) {
		oapiGetHeading(GetHandle(),&PadHeading);
		PadHeading *= 180.0/PI;
		PadHeading += 180.0;
		Guidance (true);
		SetThrusterGroupLevel(THGROUP_MAIN,1.0);
	}

	if (Firstsweep) {
		Firstsweep = false;
		Secondsweep = true;
		if (SpinTable != 0) {
			VESSELSTATUS vs;
			GetStatus(vs);
			vs.status = 0;
			vs.vrot = _V(0,0,SpinTable);
			DefSetState(&vs);
			SpinTable = 0;
		}

	//  Get the first attachment point, and duplicate it for secondary payloads.
		GetAttachmentParams (GetAttachmentHandle (false, 0), pos, dir, rot);
		//		     			   child       pos           dir       rot      handle
		SecPayload = CreateAttachment (false, pos, dir, rot, "X");


		N=14;
		double *fenergy = new double[N];
		double *fcpitch = new double[N];

		fenergy[0]				=    0.0;
		fenergy[1]				=    0.4;
		fenergy[2]				=    0.5;
		fenergy[3]				=    0.8;
		fenergy[4]				=    1.2;
		fenergy[5]				=    1.7;
		fenergy[6]				=    2.8;
		fenergy[7]				=    4.0;
		fenergy[8]				=    7.0;
		fenergy[9]				=   10.0;
		fenergy[10]				=   16.0;
		fenergy[11]				=   22.0;
		fenergy[12]				=   28.0;
		fenergy[13]				=   40.0;

		fcpitch[0]				= 89.0;
		fcpitch[1]				= 80.0;
		fcpitch[2]				= 70.0;
		fcpitch[3]				= 55.0;
		fcpitch[4]				= 48.0;
		fcpitch[5]				= 43.0;
		fcpitch[6]				= 35.0;
		fcpitch[7]				= 24.0;
		fcpitch[8]				= 18.0;
		fcpitch[9]				= 12.0;
		fcpitch[10]				=  7.0;
		fcpitch[11]				=  5.0;
		fcpitch[12]				=  2.5;
		fcpitch[13]				=  0.0;
		ProgramPitchMet( N, fenergy, fcpitch );

	}
}

void VELCRO::GetDocks()
{
		if (mode == 1) {			
//			AtchV->DetachChild (tgtatch, 0.0);
//			AttachChild (hV, attachpos[chosenpoint], tgtatch);
			numdocks = AtchV->DockCount();
																			// Find the docking ports on our attached vessel,
			VECTOR3 global, pos, dir, rot;									// and add them to our own!
			for (int i = 0 ; i < AtchV->DockCount() ; i++ ) {
				dh[i] = AtchV->GetDockHandle(i);
				AtchV->GetDockParams(dh[i],dpos[i], ddir[i], drot[i]);
				AtchV->Local2Global (dpos[i],global);
				Global2Local(global,pos);

				AtchV->HorizonRot (ddir[i],global);
				HorizonInvRot(global,dir);

				AtchV->HorizonRot (drot[i],global);
				HorizonInvRot(global,rot);

				AtchV->Undock(i);
				AtchV->DelDock(dh[i]);

				CreateDock(pos+_V(0,0,0.0),dir,rot);
			}
		}

}

void VELCRO::GiveDocks()
{
			double originalnumdocks = DockCount() - numdocks;
			for (int i = 0 ; i < numdocks; i++ ) {
//				DelDock(GetDockHandle(i+originalnumdocks));
				dh[i] = AtchV->CreateDock(dpos[i],ddir[i],drot[i]);
			}
			Undock(ALLDOCKS);

}

void VELCRO::VectorThrust()
{

	if (mode > 1) return;

	if (RCSThrust == 0.0)
		return;

	if (Firstsweep)
		return;

	if (stage > 0)
		return;

	if (DoEnginesLater > 0)
		return;

	if (oapiGetTimeAcceleration() > 100) return;

	if (mode < 2) {
		if (GetFocus() == GetHandle()) {
			tempP = GetManualControlLevel(THGROUP_ATT_PITCHDOWN, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
				GetManualControlLevel(THGROUP_ATT_PITCHUP, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

			tempY = GetManualControlLevel(THGROUP_ATT_YAWLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
				GetManualControlLevel(THGROUP_ATT_YAWRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

			tempR = GetManualControlLevel(THGROUP_ATT_BANKLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
				GetManualControlLevel(THGROUP_ATT_BANKRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

		} else {
			tempP = 0.0;
			tempY = 0.0;
			tempR = 0.0;
		}
		GetAngularVel(AngularVel);
	}
	else {
		if ((GetFocus() == GetHandle()) || (GetFocus() == hV)) {
			tempP = AtchV->GetManualControlLevel(THGROUP_ATT_PITCHDOWN, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
				AtchV->GetManualControlLevel(THGROUP_ATT_PITCHUP, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

			tempY = AtchV->GetManualControlLevel(THGROUP_ATT_YAWLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
				AtchV->GetManualControlLevel(THGROUP_ATT_YAWRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

			tempR = AtchV->GetManualControlLevel(THGROUP_ATT_BANKLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
				AtchV->GetManualControlLevel(THGROUP_ATT_BANKRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE);

		} else {
			tempP = 0.0;
			tempY = 0.0;
			tempR = 0.0;
		}
		AtchV->GetAngularVel(AngularVel);
	}

	tempP+=(2*AngularVel.x);			// Bias the thrust vector AGAINST the angular velocity a bit
	tempY+=(-2*AngularVel.y);			// This will damp angular velocity, preventing loss of control
	tempR+=(2*AngularVel.z);


//		double temppitchinput = pitchinput;
//		double tempyawinput = yawinput;
//		HorizonInvRot (_V(tempyawinput, temppitchinput, 0),_V(yawinput, pitchinput,0));
//		pitchinput = -pitchinput*cos(GetBank())-yawinput*sin(GetBank());
//		yawinput *= -yawinput*cos(GetBank())-temppitchinput*sin(GetBank());

		if (guidance) {
			tempP+=pitchinput;
			tempY+=yawinput;
			tempR+=rollinput;
		}

	if (oapiGetTimeAcceleration() > 10) {
		tempP *= 0.05;
		tempY *= 0.01;
		tempR *= 0.05;
	}
//		SetAttitudeRotLevel(_V(pitchinput,yawinput,rollinput));
	if (Smoldering && !MECO_Cond && !PostMECO)
		SetAttitudeRotLevel(_V(-tempP,tempY,-tempR));
	else SetAttitudeRotLevel(_V(0,0,-tempR));

	if (NegativeSteer) {
		tempP *= -1.0;
		tempY *= -1.0;
		}

/*	if (mode < 2) {
		SetThrusterDir(th_first[0],_V(0.012*tempY,0.012*(tempP),1));
	} else {
		AtchV->SetThrusterDir(th_first[0],_V(0.012*tempY,0.012*(tempP),1));
	}
*/
}

void VELCRO::FairingJettison (int numfairings, VECTOR3 InitialJettVector)
{
	VESSEL *DepV;
	VECTOR3 rot;
//	double throttlesetting = GetThrusterGroupLevel(THGROUP_MAIN);
	double JettMagnitude = sqrt(InitialJettVector.x*InitialJettVector.x+InitialJettVector.y*InitialJettVector.y);
	double JettInitAngle = atan(InitialJettVector.x/InitialJettVector.y);
	for (int i = 0 ; i < numfairings ; i++ ) {
//		OverridePayloadVelocity(_V(InitialJettVector.x*cos(i*(6.28/numfairings))+InitialJettVector.y*sin(i*(6.28/numfairings)),InitialJettVector.y*cos(i*(6.28/numfairings))+InitialJettVector.x*sin(i*(6.28/numfairings)),InitialJettVector.z));
		OverridePayloadVelocity(_V(-JettMagnitude*cos(JettInitAngle+(i*(6.28/numfairings))),-JettMagnitude*sin(JettInitAngle+(i*(6.28/numfairings))),InitialJettVector.z));
		DepV = DeployPayload();
		rot.z = 0.0;
		rot.x = -JettMagnitude*cos(JettInitAngle+(i*(6.28/numfairings)));
		rot.y = -JettMagnitude*sin(JettInitAngle+(i*(6.28/numfairings)));
//		rot.x = InitialJettVector.x*cos(i*(6.28/numfairings))+InitialJettVector.y*sin(i*(6.28/numfairings));
//		rot.y = InitialJettVector.y*cos(i*(6.28/numfairings))+InitialJettVector.x*sin(i*(6.28/numfairings));
		DepV->AddForce (_V(10*DepV->GetMass()*rot.x, 10*DepV->GetMass()*rot.y, 0), _V(0, 0, DepV->GetSize()/20));
	}
//	SetMainStage();
//	SetThrusterGroupLevel(THGROUP_MAIN,throttlesetting);
	ReInit = true;
}

/* CVE-lite function ! */
/* IMPORTANT */
/* VisualCreated() is a function passed through from CVE-lite as a direct replacement
/* for the DLLCLBK ovcVisualCreated() function.
/* You can't use ovcVisualCreated() with CVE-lite, so use this instead! */

void VELCRO::VisualCreated ( VESSEL *vessel, VISHANDLE vis, int refcount) {
	/* local equivalent to the DLLCLBK ovcVisualCreated() */
	SetMainStage(); // this is a good place to ask for the initial setup.
}


void VELCRO::ScanLines(FILEHANDLE cfg, void *vs)
{
	int i = 0;
	char* pLine;
	while(oapiReadScenario_nextline(cfg, pLine))
	{
		if(!strnicmp(pLine, "Meshname", 8))
		{
			char* pLine2 = pLine + 8;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%31s", &meshname);
				hVELCRO			= oapiLoadMeshGlobal (meshname);
			}

		}
		else if(!strnicmp(pLine, "MaxMainThrust", 13))
		{
			char* pLine2 = pLine + 13;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &EngineThrust);
			}

		}
		else if(!strnicmp(pLine, "IspSL", 5))
		{
			char* pLine2 = pLine + 5;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &IspSL);
			}

		}
		else if(!strnicmp(pLine, "Isp", 3))
		{
			char* pLine2 = pLine + 3;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &IspVac);
			}

		}
		else if(!strnicmp(pLine, "MaxAttitudeThrust", 17))
		{
			char* pLine2 = pLine + 17;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &RCSThrust);
			}

		}
		else if(!strnicmp(pLine, "Mass", 4))
		{
			char* pLine2 = pLine + 4;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &Mass);
			}

		}
		else if(!strnicmp(pLine, "MaxFuel", 7))
		{
			char* pLine2 = pLine + 7;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &OriginalFuelMass);
			}

		}
		else if(!strnicmp(pLine, "PropellantResource1", 19))
		{
			char* pLine2 = pLine + 19;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &OriginalFuelMass);
			}

		}
		else if(!strnicmp(pLine, "MEngineVec", 10))
		{
			char* pLine2 = pLine + 11;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf%lf%lf", &MEngineVec.x,&MEngineVec.y,&MEngineVec.z);
			}

		}
		else if(!strnicmp(pLine, "ExhaustType", 11))
		{
			char* pLine2 = pLine + 12;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				char ExhStr[31];
				pLine2++;
				sscanf(pLine2, "%31s", &ExhStr);
				numexhausts = 0;
				if(!strnicmp(ExhStr, "SRM", 3))
					ExhaustType = SRM;
				else if(!strnicmp(ExhStr, "N2O4", 4))
					ExhaustType = N2O4;
				else if(!strnicmp(ExhStr, "LF2", 3))
					ExhaustType = LF2;
				else if(!strnicmp(ExhStr, "LH2", 3))
					ExhaustType = LH2;
				else if(!strnicmp(ExhStr, "KERO", 4))
					ExhaustType = KERO;			
				else if(!strnicmp(ExhStr, "HYBRID", 6))
					ExhaustType = HYBRID;			
			}

		}
		else if(!strnicmp(pLine, "MEngineRef", 10))
		{
			char* pLine2 = pLine + 11;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
//				sscanf(pLine2, "%lf", &i);
				pLine2++;
				sscanf(pLine2, "%lf%lf%lf", &MEngineRef[numexhausts].x,&MEngineRef[numexhausts].y,&MEngineRef[numexhausts].z);
				numexhausts++;
			}

		}
		else if(!strnicmp(pLine, "Wing", 4))
		{
			char* pLine2 = pLine + 4;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf%lf%lf%lf%lf%lf", &Wing.Location.x,&Wing.Location.y,&Wing.Location.z,&Wing.span,&Wing.S,&Wing.AR);
			}

		}
		else if(!strnicmp(pLine, "HStab", 5))
		{
			char* pLine2 = pLine + 5;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf%lf%lf%lf%lf%lf", &HStab.Location.x,&HStab.Location.y,&HStab.Location.z,&HStab.span,&HStab.S,&HStab.AR);
			}

		}
		else if(!strnicmp(pLine, "VStab", 5))
		{
			char* pLine2 = pLine + 5;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf%lf%lf%lf%lf%lf", &VStab.Location.x,&VStab.Location.y,&VStab.Location.z,&VStab.span,&VStab.S,&VStab.AR);
			}

		}
		else if(!strnicmp(pLine, "Become", 6))
		{
			char* pLine2 = pLine + 6;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%31s", &becomename);
				BecomeSomethingElse = true;
			}

		}
		else if(!strnicmp(pLine, "Mayfly", 6))
		{
			char* pLine2 = pLine + 6;
			while(*pLine2 != '=' && *pLine2 != '\0')
			{
				pLine2++;
			}
			if(*pLine2 != '\0')
			{
				pLine2++;
				sscanf(pLine2, "%lf", &mayfly);
			}

		}
		/* CVE-lite function ! */
		/* IMPORTANT */
		/* Call this for every pLine in the scenario file. It will automatically
		/* deduce which pLines are for it, and load the payloads from the file. */

		/* load payload pLines */
		else if (!strnicmp (pLine, "CONFIGURATION", 13)) {
		    sscanf (pLine+13, "%d", &stage);
		} else if (!strnicmp (pLine, "HEADING", 7)) {
			sscanf(pLine+7, "%lf", &initialheading);
		} else if (!strnicmp (pLine, "DELAYSTART", 10)) {
			sscanf(pLine+10, "%lf", &DelayStart);
		} else if (!strnicmp (pLine, "PRIMEBOOSTER", 12)) {
			sscanf(pLine+12, "%d", &PrimeBooster);
		} else if (!strnicmp (pLine, "GETDOCKINGPORTS", 15)) {
			sscanf(pLine+15, "%d", &GetDockingPorts);
		} else if (!strnicmp (pLine, "NO_JETT", 7)) {
			sscanf(pLine+7, "%d", &NoJett);
		} else if (!strnicmp (pLine, "IGNITE_WITH_CORE", 16)) {
			sscanf(pLine+16, "%d", &IgniteWithCore);
		} else if (!strnicmp (pLine, "JETT_ON_MECO", 12)) {
			sscanf(pLine+12, "%d", &JettOnMECO);
		} else if (!strnicmp (pLine, "JETT_ON_EMPTY", 13)) {
			sscanf(pLine+13, "%d", &JettOnEmpty);
		} else if (!strnicmp (pLine, "IGNITENEXT", 10)) {
			sscanf(pLine+10, "%d", &IgniteNext);
		} else if (!strnicmp (pLine, "CENTERTHRUST", 12)) {
			sscanf(pLine+12, "%d", &CenterThrust);
		} else if (!strnicmp (pLine, "COUNTINGDOWN", 12)) {
			sscanf(pLine+12, "%d", &CountingDown);
		} else if (!strnicmp (pLine, "PADBIAS", 7)) {
			sscanf(pLine+7, "%lf", &padbias);
		} else if (!strnicmp (pLine, "THROTTLEBACK", 12)) {
			sscanf(pLine+12, "%lf", &ThrottleBack);
		} else if (!strnicmp (pLine, "TIPOFF", 6)) {
			sscanf(pLine+6, "%lf", &Tipoff);
		} else if (!strnicmp (pLine, "SEPFORCE", 8)) {
			sscanf(pLine+8, "%lf", &SepForce);
		} else if (!strnicmp (pLine, "SPINTABLE", 9)) {
			sscanf(pLine+9, "%lf", &SpinTable);
		} else if (!strnicmp (pLine, "ROLLATT", 7)) {
			sscanf(pLine+7, "%d", &RollAtt);
		} else if (!strnicmp (pLine, "SCREENHEIGHT", 12)) {
			sscanf(pLine+12, "%lf", &ScreenHeight);
		} else if (!strnicmp (pLine, "PITCHMULTIPLE", 13)) {
			sscanf(pLine+13, "%lf", &PitchMult);
		} else if (!strnicmp (pLine, "MINPITCH", 8)) {
			sscanf(pLine+8, "%lf", &MinPitch);
		} else if (!strnicmp (pLine, "GUIDANCE", 8)) {
			sscanf(pLine+8, "%d", &Guid);
		} else if (!strnicmp (pLine, "SHIFTFOCUS", 10)) {
			sscanf(pLine+10, "%d", &ShiftFocus);
		} else if (!strnicmp (pLine, "CAMERA", 6)) {
			sscanf(pLine+6, "%lf%lf%lf", &CAMERA_OFS.x, &CAMERA_OFS.y, &CAMERA_OFS.z);
		} else if (!strnicmp (pLine, "SHIFTPARENTCOM", 14)) {
			sscanf(pLine+14, "%lf%lf%lf", &ShiftParentCOM.x, &ShiftParentCOM.y, &ShiftParentCOM.z);
		} else if (!strnicmp (pLine, "FAIRING", 7)) {
			sscanf(pLine+7, "%d%lf%lf%lf%lf", &FairingNum, &FairingJettPress, &FairingJett.x, &FairingJett.y, &FairingJett.z );
		} else if (!strnicmp (pLine, "SLA", 3)) {
			sscanf(pLine+3, "%d%lf%lf%lf", &SLANum, &SLAJett.x, &SLAJett.y, &SLAJett.z );
		} else if (!strnicmp (pLine, "LES", 3)) {
			sscanf(pLine+3, "%lf", &LEStime );
//		} else if (!strnicmp (pLine, "ATTACHED", 8)) {
//		sscanf (pLine+8, "%s", &null);	
		} else if (!strnicmp (pLine, "PAYATCH", 7)) {
		sscanf (pLine+7, "%31s%lf%lf%lf%lf%lf%lf%lf%lf%lf", tgtname2, &tgtpos2.x,&tgtpos2.y,&tgtpos2.z,&tgtdir2.x,&tgtdir2.y,&tgtdir2.z,&tgtrot2.x,&tgtrot2.y,&tgtrot2.z);
		if ((tgtrot2.x == 0.0) && (tgtrot2.y == 0.0) && (tgtrot2.z == 0.0)) {
			tgtrot2 = tgtdir2;
			tgtdir2 = _V(0,0,-1);
		}
		PayAtch = true;
		} else if (!strnicmp (pLine, "CROSSFEED", 9)) {
		sscanf (pLine+9, "%d%31s%lf%d%lf%lf%lf%lf%lf%lf%lf%lf%lf", &chosenpoint,tgtname, &FeedRate, &PassToTank, &tgtpos.x,&tgtpos.y,&tgtpos.z,&tgtdir.x,&tgtdir.y,&tgtdir.z,&tgtrot.x,&tgtrot.y,&tgtrot.z);
		if ((tgtrot.x == 0.0) && (tgtrot.y == 0.0) && (tgtrot.z == 0.0)) {
			tgtrot = tgtdir;
			tgtdir = _V(0,0,-1);
		}
		mode = 4;
		} else if (!strnicmp (pLine, "BOOSTERBURN", 11)) {
		sscanf (pLine+12, "%d%31s%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf", &chosenpoint,tgtname, &endlevel, &tgtpos.x,&tgtpos.y,&tgtpos.z,&tgtdir.x,&tgtdir.y,&tgtdir.z,&tgtrot.x,&tgtrot.y,&tgtrot.z);
		if ((tgtrot.x == 0.0) && (tgtrot.y == 0.0) && (tgtrot.z == 0.0)) {
			tgtrot = tgtdir;
			tgtdir = _V(0,0,-1);
		}
		mode = 3;
		} else if (!strnicmp (pLine, "PARALLELBURN", 12)) {
		sscanf (pLine+12, "%d%31s%lf%lf%lf%lf%lf%lf%lf%lf%lf", &chosenpoint,tgtname, &tgtpos.x,&tgtpos.y,&tgtpos.z,&tgtdir.x,&tgtdir.y,&tgtdir.z,&tgtrot.x,&tgtrot.y,&tgtrot.z);
		if ((tgtrot.x == 0.0) && (tgtrot.y == 0.0) && (tgtrot.z == 0.0)) {
			tgtrot = tgtdir;
			tgtdir = _V(0,0,-1);
		}
		mode = 2;
		} else if (!strnicmp (pLine, "SERIESBURN", 10)) {
		sscanf (pLine+10, "%d%31s%lf%lf%lf%lf%lf%lf%lf%lf%lf", &chosenpoint,tgtname, &tgtpos.x,&tgtpos.y,&tgtpos.z,&tgtdir.x,&tgtdir.y,&tgtdir.z,&tgtrot.x,&tgtrot.y,&tgtrot.z);
		if ((tgtrot.x == 0.0) && (tgtrot.y == 0.0) && (tgtrot.z == 0.0)) {
			tgtrot = tgtdir;
			tgtdir = _V(0,0,-1);
		}
		mode = 1;
		} else if( !strncmp( pLine, "TGT_HEADING", 11 ) ) {
            sscanf (pLine+11, "%lf", &d_heading);
			PadHeading = d_heading + 180.0;  // It's not pad heading any more; just target heading
			INCLINATION = acos(sin(PadHeading*PI/180)*cos(Latitude))*DEG;
			d_heading = PadHeading;

		} else ParsePayloadString(pLine);

		if (vs != NULL)
			ParseScenarioLineEx (pLine, vs);

	}
}
void VELCRO::SetClassCaps(  FILEHANDLE cfg )
{

	ScanLines(cfg, NULL);
	CheckCentralMessage();
	if (Mass == -1)
		Mass = GetEmptyMass();
	SetEmptyMass ( Mass  );

	SetMainStage();
}

void VELCRO::SaveState( FILEHANDLE scn )
{
//	if (mode > 1) 
//		if ((!Secondsweep) && (!Firstsweep) || (!PrimeBooster))
//		AtchV->DetachChild (tgtatch, 0.0);

	char cbuf[256];

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Call this after the normal SaveDefautState() to record attached payloads
	/* it the scenario file */





	if ((tempatch != NULL) && (!EnginesStarted))
			DetachChild (attachpos[chosenpoint]);
	else {
		if (mode == 1) 
			DetachChild (attachpos[chosenpoint], 2.0);
		if (mode > 1)
			AtchV->DetachChild (tgtatch, 2.0);
	}
	if (GetDocksDone)
		Undock(ALLDOCKS);



	SaveDefaultState (scn);

	oapiWriteScenario_int (scn, "CONFIGURATION", stage);
	oapiWriteScenario_int (scn, "PRIMEBOOSTER", PrimeBooster);
	oapiWriteScenario_int (scn, "CENTERTHRUST", CenterThrust);
	if (RollAtt != 0)
		oapiWriteScenario_int (scn, "ROLLATT", RollAtt);
	if (!IgniteNext)
		oapiWriteScenario_int (scn, "IGNITENEXT", IgniteNext);
	if (ThrottleBack != 99999.9)
		oapiWriteScenario_float (scn, "THROTTLEBACK", ThrottleBack);
	if (SpinTable != 0.0)
		oapiWriteScenario_float (scn, "SPINTABLE", SpinTable);
	if (NoJett == true)
		oapiWriteScenario_int (scn, "NO_JETT", NoJett);
	if (CountingDown)
		oapiWriteScenario_int (scn, "COUNTINGDOWN", CountingDown);
	if (GetDockingPorts)
		oapiWriteScenario_int (scn, "GETDOCKINGPORTS", GetDockingPorts);
	if (IgniteWithCore == false)
		oapiWriteScenario_int (scn, "IGNITE_WITH_CORE", IgniteWithCore);
	if (JettOnMECO == false)
		oapiWriteScenario_int (scn, "JETT_ON_MECO", JettOnMECO);
	if (JettOnEmpty == 0)
		oapiWriteScenario_int (scn, "JETT_ON_EMPTY", JettOnEmpty);
	if (DelayStart > 0)
		oapiWriteScenario_float (scn, "DELAYSTART", DelayStart);
	if (padbias =! 0.0)
		oapiWriteScenario_float (scn, "PADBIAS", padbias);
	if (ScreenHeight != 125.0)
		oapiWriteScenario_float (scn, "SCREENHEIGHT", ScreenHeight);
	if (Tipoff != 1.0)
		oapiWriteScenario_float (scn, "TIPOFF", Tipoff);
	if (SepForce != 1.0)
		oapiWriteScenario_float (scn, "SEPFORCE", SepForce);
	if (PitchMult != 1.0)
		oapiWriteScenario_float (scn, "PITCHMULTIPLE", PitchMult);
	if (MinPitch != 0.0)
		oapiWriteScenario_float (scn, "MINPITCH", MinPitch);
	if (guidance != 0)
		oapiWriteScenario_int (scn, "GUIDANCE", guidance);
	if (ShiftFocus != 1)
		oapiWriteScenario_int (scn, "SHIFTFOCUS", ShiftFocus);
	if (CAMERA_OFS.x != -100)
		oapiWriteScenario_vec (scn, "CAMERA", CAMERA_OFS);
	if ((ShiftParentCOM.x != 0.0) || (ShiftParentCOM.y != 0.0) || (ShiftParentCOM.z != 0.0))
		oapiWriteScenario_vec (scn, "SHIFTPARENTCOM", ShiftParentCOM);
	if (FairingNum > 0) {
		sprintf (cbuf, "%d %0.2f %0.2f %0.2f %0.2f", FairingNum, FairingJettPress, FairingJett.x, FairingJett.y, FairingJett.z);
		oapiWriteScenario_string (scn, "FAIRING", cbuf);
	}
	if (SLANum > 0) {
		sprintf (cbuf, "%d %0.2f %0.2f %0.2f", SLANum, SLAJett.x, SLAJett.y, SLAJett.z);
		oapiWriteScenario_string (scn, "SLA", cbuf);
	}
	if (LEStime > 0)
		oapiWriteScenario_float (scn, "LES", LEStime);

	oapiWriteScenario_float (scn, "TGT_HEADING", PadHeading - 180.0);
	if (stage == 0) {
		if (mode == 1) {
			sprintf (cbuf, "%d %s %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f", chosenpoint, tgtname, tgtpos.x, tgtpos.y, tgtpos.z, tgtdir.x, tgtdir.y, tgtdir.z, tgtrot.x, tgtrot.y, tgtrot.z);
			oapiWriteScenario_string (scn, "SERIESBURN", cbuf);
		}
		else if (mode == 2) {
			sprintf (cbuf, "%d %s %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f", chosenpoint, tgtname, tgtpos.x, tgtpos.y, tgtpos.z, tgtdir.x, tgtdir.y, tgtdir.z, tgtrot.x, tgtrot.y, tgtrot.z);
			oapiWriteScenario_string (scn, "PARALLELBURN", cbuf);
		}
		else if (mode == 3) {
			sprintf (cbuf, "%d %s %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f", chosenpoint, tgtname, endlevel, tgtpos.x, tgtpos.y, tgtpos.z, tgtdir.x, tgtdir.y, tgtdir.z, tgtrot.x, tgtrot.y, tgtrot.z);
			oapiWriteScenario_string (scn, "BOOSTERBURN", cbuf);
		}
		else if (mode == 4) {
			sprintf (cbuf, "%d %s %0.4f %d %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f", chosenpoint, tgtname, FeedRate, PassToTank, tgtpos.x, tgtpos.y, tgtpos.z, tgtdir.x, tgtdir.y, tgtdir.z, tgtrot.x, tgtrot.y, tgtrot.z);
			oapiWriteScenario_string (scn, "CROSSFEED", cbuf);
		}

	}


	/* save payload lines */
	SaveDefaultStateEx(scn);

// Stitch these back together,
// just in case we're saving due to a Record event....
//

	if (stage == 0) {
		if ((tempatch != NULL) && (!EnginesStarted))
			AttachChild (hV, attachpos[chosenpoint], tempatch);

		else {
			if (mode == 1) 
				AttachChild (hV, attachpos[chosenpoint], tgtatch);
			if (mode > 1)
				AtchV->AttachChild (MeV, tgtatch, attachpos[chosenpoint+2]);
		}
	}
}

void VELCRO::LoadState(  FILEHANDLE scn, void *vs )
{
//    char *line;
//	char null[31];

	ScanLines(scn, vs);





	/* always set the stage whenever something may have changed */
	SetMainStage();

	if (padbias != 0.0) {
		VECTOR3 currTD1, currTD2, currTD3;
		GetTouchdownPoints(currTD1, currTD2, currTD3);
		SetTouchdownPoints(_V(currTD1.x,currTD1.y,currTD1.z+padbias), _V(currTD2.x,currTD2.y,currTD2.z+padbias), _V(currTD3.x,currTD3.y,currTD3.z+padbias));
	}


}


void VELCRO::PlaceComponents()
{
	COMMON_MESH tmp;

	tmp.MeshHandle(hVELCRO);
	tmp.MeshPosition(VELCRO_OFS);
	AddMeshEx (tmp);

}

/* IMPORTANT */
/* Always have a function that you can call as many times as you like that will re-draw
/* all you meshes, reset all your thrusters and check to make sure the propellent is
/* there. This should be called after any configuration change, or any GDPS line (see
/* TimeStep(), above) */

void VELCRO::SetMainStage( )
{

	SetVelcroStage();
}

void VELCRO::SetVelcroStage( )
{

	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	if (stage == 1) {
		atchmass = 0;
	}
	if (Mass != -1.0)
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
		SetEmptyMass ( Mass + atchmass + secmass + CalculateMass() );


		SetPitchMomentScale (1e-6);
		SetBankMomentScale (1e-6);

	if (CAMERA_OFS.x != -100)
		SetCameraOffset(CAMERA_OFS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	ClearAirfoilDefinitions();

	if (Wing.AR > 0.0)
		CreateAirfoil (LIFT_VERTICAL, Wing.Location, VLiftCoeff, Wing.span, Wing.S, Wing.AR);
	if (HStab.AR > 0.0)
		CreateAirfoil (LIFT_VERTICAL, HStab.Location, VLiftCoeff2, HStab.span, HStab.S, HStab.AR);
	if (VStab.AR > 0.0)
		CreateAirfoil (LIFT_HORIZONTAL, VStab.Location, HLiftCoeff, VStab.span, VStab.S, VStab.AR);


	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hVELCRO, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */

	if ((ph_first==NULL) && (stage == 0)) {
		ClearPropellantResources();
		ph_first  = CreatePropellantResource(OriginalFuelMass,OriginalFuelMass*InitFuelLevel,1.0); 
	}
	else if (ph_first==NULL) {
		ClearPropellantResources();
		ph_first  = CreatePropellantResource(OriginalFuelMass,fuellevel,1.0); 
	}
	SetDefaultPropellantResource (ph_first); 

	ClearThrusterDefinitions();

	if (IspSL < 0.0)
		IspSL = IspVac;



		th_first[0] = CreateThruster ( _V(0,0,MEngineRef[0].z), MEngineVec, EngineThrust, ph_first, IspVac, IspSL);
		if (MEngineRef[0].z > 0)
			NegativeSteer = true;

		thg_first = CreateThrusterGroup (th_first, 1, THGROUP_MAIN);
	
		numengines = 1;

	MakeExhausts();

	SetAttControlsVELCRO();
}



void VELCRO::SetAttachedStage( )
{

	SetVelcroStage();

	if ((Secondsweep) || (Firstsweep)) return;

	ClearThrusterDefinitions();
	ClearPropellantResources();


	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */

	totalemptymass = Mass + AtchV->GetEmptyMass() + CalculateMass();
	oapiSetEmptyMass (hV,totalemptymass);
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */

	VECTOR3 currTD1, currTD2, currTD3;
	GetTouchdownPoints(currTD1, currTD2, currTD3);
	if (PrimeBooster)
		AtchV->SetTouchdownPoints (_V(currTD1.x,currTD1.y-0.01,currTD1.z-0.1-attach[chosenpoint].z+tgtpos.z+padbias), _V(currTD2.x,currTD2.y-0.01,currTD2.z-0.1-attach[chosenpoint].z+tgtpos.z+padbias), _V(currTD3.x,currTD3.y-0.01,currTD3.z-0.1-attach[chosenpoint].z+tgtpos.z+padbias));



	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */
	ph_first  = AtchV->CreatePropellantResource(OriginalFuelMass,OriginalFuelMass*InitFuelLevel,1.0);  
	SetDefaultPropellantResource (ph_first); 

	numengines = 1;
	DoEnginesLater = 1;


}


void VELCRO::MakeExhausts()
{



//	th_first[0] = GetThrusterHandleByIndex(0);
	VECTOR3 dir = _V(-1*MEngineVec.x,-1*MEngineVec.y,-1*MEngineVec.z);

	PARTICLESTREAMSPEC contrail = 
	{
		0, pow(GetThrusterMax(th_first[0])/(numexhausts+1),0.3333)*0.08, 24, 250, 0.4, 2.0, 7, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, pow(GetThrusterMax(th_first[0])/(numexhausts+1),0.3333)*0.04, 45, 250, 0.2, 0.4, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.3, 0.3
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");


	if (numexhausts > 0)
	{
		ClearExhaustRefs();

		for (int i = 0 ; i < numexhausts; i++ ) {
			if (ExhaustType == N2O4)
				AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.5,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,MEngineRef[i],dir,texstor);
			else if (ExhaustType == LH2)
				AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.5,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,MEngineRef[i],dir,tex);
			else if (ExhaustType == KERO)
				AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.75,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.055,MEngineRef[i],dir,texkero);
			else if (ExhaustType == LF2)
				AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.5,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,MEngineRef[i],dir,texfluor);
			else if ((ExhaustType == SRM) || (ExhaustType == HYBRID)) {
				AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.08,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,MEngineRef[i],dir,texstor);
				AddExhaustStream (th_first[0], MEngineRef[i], &contrail);
				AddExhaustStream (th_first[0], MEngineRef[i], &exhaust_main);
			}
		}
	}

/*	for (int i = 0 ; i < GetThrusterCount(); i++ ) {
		th_first[i] = GetThrusterHandleByIndex(i);
		AddExhaust(th_first[i],GetThrusterMax(th_first[i])*0.0002,GetThrusterMax(th_first[i])*0.000012,tex);
	}
*/

//	if (ph_first==NULL)  
//		ph_first  = CreatePropellantResource(FIRST_PROP_MASS,FIRST_PROP_MASS*InitFuelLevel,1.0); 



//	ph_main = NULL;
//	ph_main = GetPropellantHandleByIndex(0);
//	SetPropellantEfficiency(GetPropellantHandleByIndex(0),1.0);
//	double currmass = GetFuelMass();
//	double maxmass = GetMaxFuelMass();
//	ClearPropellantResources();
//	ph_main = CreatePropellantResource(maxmass,currmass,1.0);
//	SetDefaultPropellantResource (ph_main); 
	
//	th_first[0] = CreateThruster ( _V(0,0.72,-4), _V(0,0,1), THRUST_FIRST_VAC, ph_first, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &exhaust_main);

//	th_first[1] = CreateThruster ( _V(0,-0.72,-4), _V(0,0,1), THRUST_FIRST_VAC, ph_first, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &exhaust_main);

//	thg_first = CreateThrusterGroup (th_first, 2, THGROUP_MAIN);
	
//	AddExhaust (th_first[0], 8.0, 0.9, tex);
//	AddExhaust (th_first[1], 8.0, 0.9, tex);

}

void VELCRO::DoEngines( )
{


	DoEnginesLater = 0;
	EnginesDone = true;

	VECTOR3 local, global, newlocal;
	VECTOR3 dir; // = _V(-1*MEngineVec.x,-1*MEngineVec.y,-1*MEngineVec.z);

	local = _V(0,0,MEngineRef[0].z);
	if (CenterThrust)
		newlocal = local;
	else {
		Local2Global (local,global);
		AtchV->Global2Local(global,newlocal);
	}

//	VECTOR3 dirl = MEngineVec; //_V(-1*MEngineVec.x,-1*MEngineVec.y,-1*MEngineVec.z);
	HorizonRot (MEngineVec,global);
	AtchV->HorizonInvRot(global,dir);
	 
	if (mode == 3) {
		for (i = 0 ; i < AtchV->GetPropellantCount(); i++ ) {
			if (AtchV->GetPropellantMass(AtchV->GetPropellantHandleByIndex(i)) == AtchV->GetFuelMass()) {   // Found the default propellant!
				if (ph_otherdef == NULL) 
					ph_otherdef = AtchV->GetPropellantHandleByIndex(i);
			}
		}
//_V(0,0,MEngineRef[0].z)
		th_first[0] = AtchV->CreateThruster ( newlocal, dir, EngineThrust, ph_otherdef, IspVac, IspSL);
	} else
		th_first[0] = AtchV->CreateThruster ( newlocal, dir, EngineThrust, ph_first, IspVac, IspSL);
//	AtchV->AddExhaustStream (th_first[0], newlocal, &contrail);
//	AtchV->AddExhaust (th_first[0], 30.0, 4.0, newlocal, _V(0,0,-1),tex);


	thg_first = AtchV->CreateThrusterGroup (th_first, 1, THGROUP_USER);
	th_dummy[0] = CreateThruster ( _V(0,0,MEngineRef[0].z), MEngineVec, 1e-12, ph_first, 1e12, 1e12);
	thg_dummy = CreateThrusterGroup (th_dummy, 1, THGROUP_MAIN);

//	VECTOR3 dir = _V(-1*MEngineVec.x,-1*MEngineVec.y,-1*MEngineVec.z);

	PARTICLESTREAMSPEC contrail = 
	{
		0, pow(GetThrusterMax(th_first[0]),0.3333)*0.08, 24, 250, 0.4, 2.0, 7, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, pow(GetThrusterMax(th_first[0]),0.3333)*0.04, 45, 250, 0.2, 0.4, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");



	for (int i = 0 ; i < numexhausts; i++ ) {
		Local2Global (MEngineRef[i],global);
		AtchV->Global2Local(global,newlocal);
		if (ExhaustType == N2O4)
			Exhaust[i] = AtchV->AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.5,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,newlocal,_V(-1*dir.x,-1*dir.y,-1*dir.z),texstor);
		else if (ExhaustType == LH2)
			Exhaust[i] = AtchV->AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.5,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,newlocal,_V(-1*dir.x,-1*dir.y,-1*dir.z),tex);
		else if (ExhaustType == KERO)
			Exhaust[i] = AtchV->AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.75,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.055,newlocal,_V(-1*dir.x,-1*dir.y,-1*dir.z),texkero);
		else if (ExhaustType == LF2)
			Exhaust[i] = AtchV->AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.5,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,newlocal,_V(-1*dir.x,-1*dir.y,-1*dir.z),texfluor);
		else if ((ExhaustType == SRM) || (ExhaustType == HYBRID)) {
			Exhaust[i] = AtchV->AddExhaust(th_first[0],pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.08,pow(GetThrusterMax(th_first[0])/numexhausts,0.3333)*0.035,newlocal,_V(-1*dir.x,-1*dir.y,-1*dir.z),texstor);
			AtchV->AddExhaustStream (th_first[0], newlocal, &contrail);
			AtchV->AddExhaustStream (th_first[0], newlocal, &exhaust_main);
		}
	}


}


/** Handle buffered key events sent to the vessel by orbiter. 
 *  @param dwKey The scan code of the key
 *  @param fPressed true if pressed, false if depressed
 *  @param keystate the current state of the keyboard
 *
 *  modified switch routine by dave rowbotham.
 */


void VELCRO::JettisonLES()
{
				VESSEL *depV;
//				PROPELLANT_HANDLE ph_SRM1;
//				THRUSTER_HANDLE th_sep[1];

				LEStime = 0.0;
				depV = DeployPayload(_V(0,0,100.0));
				depV->SetThrusterGroupLevel(THGROUP_MAIN,1.0);
/*				PARTICLESTREAMSPEC contrail = 
				{
			//		0, 0.25, 10.0, 9.0, 0.5, 1.0, 10, 3.0, PARTICLESTREAMSPEC::EMISSIVE,
					0, 0.75, 29, 50.0, 0.5, 0.1, 7, 3.0, PARTICLESTREAMSPEC::EMISSIVE,
					PARTICLESTREAMSPEC::LVL_SQRT, 0.1, 1,
					PARTICLESTREAMSPEC::ATM_PLOG, 0, 0
				};

				ph_SRM1  = depV->CreatePropellantResource(1,1,1.0); 
				th_sep[0] = depV->CreateThruster ( _V(0,0,0), _V( 0,0,1), 0.001, ph_SRM1, 1e3, 1e3);
				depV->AddExhaustStream (th_sep[0], _V(0,0,0), &contrail);
				depV->SetThrusterLevel(th_sep[0],1.0);
*/				
				ReInit = true;
}

void VELCRO::JettisonPart()
 {
	if ((Firstsweep) || (Secondsweep)) return;
	if (FairingNum > 0) {
		FairingJettison(FairingNum,FairingJett);
		FairingNum = 0;
	}
	/*else if ((stage == 0) && (mode > 0))
		if ((ShiftParentCOM.x != 0.0) || (ShiftParentCOM.y != 0.0) || (ShiftParentCOM.z != 0.0))
			AtchV->ShiftCG(_V(-1.0*ShiftParentCOM.x,-1.0*ShiftParentCOM.y,-1.0*ShiftParentCOM.z));
*/

	if (NoJett) return;
	if (stage == 0) {
		Undock(ALLDOCKS);
		stage = 1;
		if (GetThrusterGroupLevel(THGROUP_MAIN) > 0.01) Thrusting = true;
		SetThrusterGroupLevel(THGROUP_MAIN,0.0);
		JettisonCount = -20;
//		DelPropellantResource (ph_first); 
		PlayVesselWave3(MyID,PLAYBANG);
		OverridePayloadVelocity(_V(0,0,2.0));
		if (mode == 0) {
			AtchV = DeployPayload();
			if (AtchV == NULL) return;
		}
		switchfocus = 1;
		if (mode == 1) {
			if (GetDocksDone) GiveDocks();
			if (GetAttachmentStatus(attachpos[chosenpoint]) != NULL)
				DetachChild (attachpos[chosenpoint], 2.0);
			else {
				OverridePayloadVelocity(_V(0,0,2.0));
				DeployPayload();
			}
		}
		if (mode > 1) {
			if (ph_first != NULL)
				fuellevel = AtchV->GetPropellantMass(ph_first);
			if (AtchV->GetThrusterGroupLevel(THGROUP_MAIN) > 0.01) 
				Thrusting = true;
//			AtchV->DetachChild (tgtatch, 2.0);
			oapiSetEmptyMass (hV,originalemptymass);
//			AtchV->SetThrusterGroupLevel(thg_first,0.0);
			if (ph_first != NULL)
				AtchV->DelPropellantResource (ph_first);
			if (mode != 4)
				for (int i = 0; i < numengines; i++) {
					AtchV->DelThruster(th_first[i]);
					AtchV->DelExhaust(Exhaust[i]);
				}
			AtchV->DelThrusterGroup(thg_first,THGROUP_MAIN,false);
			AtchV->DetachChild (tgtatch, 2.0);

//			ForceNext = true;

		}

//		DeployAllPayloads();
	
	}
	else if (PayAtch) {
		PayAtch = false;
		if (GetAttachmentStatus(SecPayload) != NULL)
			DetachChild (SecPayload, 2.0);
		secmass = 0.0;
	}
	else {
		OverridePayloadVelocity(_V(0,0,2.0));
		if (!IgniteWithCore) {
			DetV = DeployPayload();
			if (DetV == NULL) return;
/*			hV=oapiGetVesselByName(AtchV->GetName());
			if (oapiGetVesselByName(GetName()) == oapiGetFocusObject())
				oapiSetFocusObject(hV);
			if ((Thrusting) && (IgniteNext))
				if (guidance)
					AtchV->SetThrusterGroupLevel(THGROUP_MAIN,GuidanceTransfer);
				else
 					AtchV->SetThrusterGroupLevel(THGROUP_MAIN,1.0);
*/			switchfocus = 1;

		} else 
			DeployPayload();
	}
	if (SLANum > 0) {
		FairingJettison(SLANum,SLAJett);
		SLANum = 0;
	}

	ReInit = true;
 }


int VELCRO::ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate)
{
	if (!fPressed) return 0; // only process keydown events
	if (KEYMOD_SHIFT (keystate)) 
	{
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		// insert ctrl key combinations here
	}
	else {

		if (KEYDOWN (keystate, OAPI_KEY_J)) 
		{ // "Jettison Stage"
			if (oapiAcceptDelayedKey (OAPI_KEY_J, 1.0))
			{
				if (FairingNum > 0) {
					FairingJettison(FairingNum,FairingJett);
					FairingNum = 0;
				}
				else JettisonPart();
						
							/* CVE-lite function ! */
							/* IMPORTANT */
							/* call DeployPayload() to jettison the GPDS payloads in sequence */


							/* IMPORTANT */
							/* Always re-set the stage after any kind of change */
//							SetMainStage();
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_O))
		{ 
			if (oapiAcceptDelayedKey (OAPI_KEY_O, 1.0))
				if (!guidance) { 
					Guidance( true );
//					guidance=true;
//					sprintf(oapiDebugString(),"Guidance ON");
					pitchinput = 0.0;
					yawinput = 0.0;
					rollinput = 0.0;
				}
				else {
					Guidance( false );
//					guidance = false;
					SetAttitudeRotLevel(_V(0,0,0));
					pitchinput = 0.0;
					yawinput = 0.0;
					rollinput = 0.0;

//					sprintf(oapiDebugString(),"Guidance OFF");
				}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_U)) 
		{ // "Set launch Azimuth"
			if (oapiAcceptDelayedKey (OAPI_KEY_U, 1.0))
			{
				GetInclinationSP();
				return 1;
			}
		}

	}
	return 0;
}




int VELCRO::ConsumeKey( const char *keystate )
{

	if (KEYMOD_SHIFT (keystate)) 
	{
						
		return 0; // shift combinations are reserved
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		return 0;
	}
	else 
	{ // unmodified keys
	
		return 0;

	}
	return 0;
}




void VELCRO::SetAttControlsVELCRO()
{
	double dist = GetSize();

	THRUSTER_HANDLE th_rot[2];

	th_rot[0] = CreateThruster (_V(0,0, dist), _V(0, 1,0), RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(0,0,-dist), _V(0,-1,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);

	th_rot[0] = CreateThruster (_V(0,0,-dist), _V(0,1,0),RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(0,0, dist), _V(0,1,0),RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_UP);

	th_rot[0] = CreateThruster (_V(0,0,-dist), _V(0,-1,0),RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(0,0, dist), _V(0,-1,0),RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_DOWN);

	th_rot[0] = CreateThruster (_V(0,0, dist), _V(0,-1,0), RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(0,0,-dist), _V(0, 1,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);

	th_rot[0] = CreateThruster (_V(0,0, dist), _V(-1,0,0), RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(0,0,-dist), _V( 1,0,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);

	th_rot[0] = CreateThruster (_V(0,0, dist), _V( 1,0,0), RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(0,0,-dist), _V(-1,0,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);

	th_rot[0] = CreateThruster (_V( dist,0.0,0.0), _V(-1,0,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_LEFT);
	th_rot[0] = CreateThruster (_V(-dist,0.0,0.0), _V( 1,0,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_RIGHT);

	th_rot[0] = CreateThruster (_V( dist*2,0,0), _V(0,-1,0), RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(-dist*2,0,0), _V(0, 1,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);

	th_rot[0] = CreateThruster (_V( dist*2,0, 0), _V(0, 1,0), RCSThrust, ph_first, IspVac);
	th_rot[1] = CreateThruster (_V(-dist*2,0,0), _V(0, -1,0), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);

	th_rot[0] = CreateThruster (_V(0,0, -dist), _V(0,0,1), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_FORWARD);
	th_rot[0] = CreateThruster (_V(0,0, dist), _V(0,0,-1), RCSThrust, ph_first, IspVac);
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_BACK);
}


