/**
*
* $Id: CVEL-VELCRO.h,v 0.1 2004/11/20 18:03:45 erv Exp $
* (c) 2004 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/


#include "../../include/cvel.h"
#include "../../include/cvelmessage.h"
#include "../../include/orbitersdk.h"
#include "../../include/OrbiterSoundSDK3.h"
#include "../../include/guidance.h"
#include "../../include/DlgCtrl.h"
#include "../../include/ScnEditorAPI.h"
#include <cstdio>
#include <cmath>
#include <string>

#define InitFuelLevel 1


#define IDD_DAMAGE                      172
#define IDD_EDITOR1                     173
#define IDD_EDITOR_PG1                  174
#define IDD_EDITOR_PG2                  175
#define IDB_SWITCH3                     175
#define IDD_EDITOR_PG3                  176


#define LOADBMP(id) (LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (id)))

// ==============================================================
// Global parameters
// ==============================================================

typedef struct {
	HINSTANCE hDLL;
	HFONT font[2];
	DWORD col[4];
	HBRUSH brush[4];
	HPEN pen[2];
} GDIParams;

GDIParams g_Param;

static HELPCONTEXT g_hc = {
	"html/vessels/deltaglider.chm",
	0,
	"html/vessels/deltaglider.chm::/deltaglider.hhc",
	"html/vessels/deltaglider.chm::/deltaglider.hhk"
};

// ==============================================================
// Local prototypes

//BOOL CALLBACK Ctrl_DlgProc (HWND, UINT, WPARAM, LPARAM);
//void UpdateCtrlDialog (DeltaGlider *dg, HWND hWnd = 0);

//BOOL CALLBACK Damage_DlgProc (HWND, UINT, WPARAM, LPARAM);
//void UpdateDamageDialog (DeltaGlider *dg, HWND hWnd = 0);

// ==============================================================
// Airfoil coefficient functions
// Return lift, moment and zero-lift drag coefficients as a
// function of angle of attack (alpha or beta)
// ==============================================================

// All data from astronautix.com or Aerospace Projects Review V1N2

const VECTOR3 VELCRO_OFS		= _V(0,0,0);
const VECTOR3 CAMERA_OFS		= _V(-0.0,4.55,20.6);

const double GuidanceTransfer = 1.0-(1e-9);		// A throttle setting of (1-very small amount) to tell the next stage to pick up the autopilot
const double Smoldering = 2e-12;				// A very very low thrust value for delay-start engines to set throttle

typedef struct {
	VECTOR3 Location;
	double span;
	double S;
	double AR;
} AIRFOIL;

#define PLAYGEAR					 20		// the gear sound
#define PLAYBANG					 24		// the jettison sound
#define PLAYJETT					 25     // entry shield jettison sound
#define PLAYJET						 26     // the jet noise sound

/* make your own vessel class (named VELCRO here, for example)
   subclassed from COMMON_VESSEL_EX */

class VELCRO : public COMMON_VESSEL_EX, GUIDANCE {
public:
	enum ExhaustType { SRM, N2O4, LF2, LH2, KERO, HYBRID } ExhaustType;
	VELCRO (OBJHANDLE hObj, int fmodel);
	~VELCRO();
	bool AzInput (void *id, char *str, void *data);
	bool IncInput (char *rstr);
	bool BecomeSomethingElse;

// Overloaded callback functions
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkLoadStateEx (FILEHANDLE scn, void *vs);
	void clbkSaveState (FILEHANDLE scn);
	void clbkPostStep (double simt, double simdt, double mjd);
	int  clbkConsumeBufferedKey (DWORD key, bool down, char *kstate);
	void clbkVisualCreated (VISHANDLE vis, int refcount);
	void clbkVisualDestroyed (VISHANDLE vis, int refcount);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);

protected:
	void TimeStep(  double simt ); // need a timestep loop to conduct the GPDS polling
	void GetDocks();
	void GiveDocks();
	void VectorThrust();
	void VisualCreated( VESSEL *vessel, VISHANDLE vis, int refcount ); // must use VisualCreated() instead of ovcVisualCreated()
	void SetClassCaps(  FILEHANDLE cfg );
	void LoadState(  FILEHANDLE scn, void *vs );
	void SaveState(  FILEHANDLE scn );
	void FairingJettison (int numfairings, VECTOR3 InitialJettVector);
	void PlaceComponents();
	void ScanLines(  FILEHANDLE cfg, void *vs );

	/* CVE-lite requirement ! */
	// Create a function that can be called as many times as you like. It is
	// vital to have a function like this that cleans out all the
	// thrusters and meshes and draws/recreates them again.
	// This is so that you can respond immediately to GPDS requests.
	// Here I've called it SetMainStage() it can be Refresh() or anything you like.
	void SetMainStage(); 
	void SetVelcroStage();
	void SetAttachedStage();
	void DoEngines();
	void MakeExhausts();
	void SetAttControlsVELCRO();    // this is to set up the thrusters. I like to keep them separate.
	void JettisonLES();
	void JettisonPart();
	int  ConsumeKey(  const char *keystate );
	int  ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate);
	double pitchinput, yawinput, rollinput;
	VESSELSTATUS vst;


	double GetCPitch(double t);
	void Lateral_Guidance(VESSEL *vessel, double SimT, double SimT_Last);
	void Guidance_Routine(VESSEL *vessel, double SimT, double SimT_Last);
	void GetInclinationSP();
	void InitialiseGuidance();

	bool Guidance() { return guidance; };
	void Guidance( bool g ) { guidance = g; };

	void DisplayInclinationSetPoint( bool d ) { DisplayIncSP=d; };
	bool DisplayInclinationSetPoint() { return DisplayIncSP; };

	double initialgamma, initialheading, launchrotate;


	PROPELLANT_HANDLE ph_first,ph_Velcro,ph_otherdef, ph_PassToTank;
	THRUSTER_HANDLE   th_first[64], th_dummy[1];
	THGROUP_HANDLE    thg_first,thg_dummy;
	MESHHANDLE		  hVELCRO;
	SURFHANDLE		  tex,texstor, texfluor, texkero, retex;
	int stage, i;
	bool havesound;
	VECTOR3 Vold, Vnew;
	double dt, oldsimt;
	int MyID;
	VECTOR3 gpos, grms, pos, dir, rot;
	OBJHANDLE hV, hvessel, MeV, detV, secV;
//	DWORD i, j;
	double atchmass, secmass;
	DOCKHANDLE hDock, Dock0, Dock1;
	bool Firstsweep, Secondsweep;
	int numengines;
	double tempP, tempY, tempR;
	VECTOR3 AngularVel;
	bool autopilot;
	bool engineignited, bgoingsouth;
	double last_simt, ignition_time, density, displaytime, simt, d_heading;
	bool first_guidance_pass;
	ATTACHMENTHANDLE  attachpos[4], tgtatch, tempatch, SecPayload, TgtSecPayload;
	VECTOR3           attach[4];
	VECTOR3			  attachdir[4];
	VECTOR3           attachrot[4];
	char tgtname[31], tgtname2[31], meshname[31], becomename[31];
	VECTOR3				tgtpos, tgtdir, tgtrot, tgtpos2, tgtdir2, tgtrot2;
	int mode, chosenpoint, chosenpoint2, switchfocus;
	VESSEL *AtchV, *DetV, *SecV;
	VECTOR3 ThrustDir;
	bool DetachMe, DetachMeDone, ForceMe, EnginesStarted, ReInit, ForceNext, EnginesDone;
	double originalemptymass,totalemptymass;
	ANIMATIONCOMPONENT_HANDLE LGr, RGr;
	int JettisonCount, DoEnginesLater;
	bool OnGround, PrimeBooster, NoJett, Thrusting;
	double Mass, OriginalFuelMass;
	double EngineThrust, IspVac, IspSL, RCSThrust;
	int numexhausts;
	int FairingNum, SLANum;
	double FairingJettPress;
	VECTOR3 FairingJett, SLAJett;
	VECTOR3 MEngineRef[64];
	VECTOR3 MEngineVec;
	VECTOR3 CAMERA_OFS;
	VECTOR3 ShiftParentCOM;
	double padbias, DelayStart, endlevel, energy, diff, yawangle, ThrottleBack, LEStime;
	bool CountingDown, CenterThrust;
	double SpinTable;
	int PassToTank;
	double Atchflowrate, FeedRate;
	VECTOR3 horiz;
	bool Flydown;
	int RollAtt;
	UINT Exhaust[512];
	double Bank_Error, Yaw_Error;
	double ScreenHeight, PitchMult, MinPitch;
	double fuellevel, Tipoff, SepForce, mayfly;
	bool NegativeSteer, IgniteWithCore, PayAtch, GetDocksDone, GetDockingPorts, JettOnMECO, IgniteNext, Guid;
	DOCKHANDLE dh[64];
	VECTOR3 dpos[64], ddir[64], drot[64];
	int numdocks, JettOnEmpty, ShiftFocus;
	AIRFOIL Wing, HStab, VStab;


};







/* basic vessel class setup */

void VELCRO::clbkSetClassCaps (FILEHANDLE cfg)
{
	SetClassCaps(cfg);
}

void VELCRO::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
	LoadState(  scn, vs );
}

void VELCRO::clbkSaveState (FILEHANDLE scn)
{
	SaveState(  scn );
}

void VELCRO::clbkPostStep (double simt, double simdt, double mjd)
{
	TimeStep(simt);
}

void VELCRO::clbkVisualCreated (VISHANDLE _vis, int refcount)
{
	SetMainStage( );
}

void VELCRO::clbkVisualDestroyed (VISHANDLE _vis, int refcount)
{
	SetMainStage( );
}

int VELCRO::clbkConsumeBufferedKey (DWORD dwKey, bool fPressed, char *keystate)
{
	int ret = ConsumeBufferedKey( dwKey, fPressed, keystate );
	return ret;
}



bool VELCRO::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
//	if (oapiCockpitMode() != COCKPIT_VIRTUAL) return false;

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	if (!MECO_Cond && !PostMECO) {

// Draw Flight Director

	double fYawError = Yaw_Error*RAD;
	double fBankError = Bank_Error;
	double fPitchError = Pitch_Error;

/*	VECTOR3 rloc = _V(0,0,0);
	HorizonInvRot(rhorizon,rloc);

	double currHdg;
	oapiGetHeading(GetHandle(),&currHdg);
	double fYawError = rhorizon.x - currHdg;
	double fPitchError = rhorizon.y - GetPitch();
	double fBankError = rhorizon.z - (GetBank()- (2*PI));
*/


	double InnCoord = 0.5;
//	if (guidance) InnCoord = 0.25;

	int iX = hps->CX + (int)(fYawError * hps->Scale);
	int iY = hps->CY - (int)(fPitchError * hps->Scale);
	int iR = hps->Markersize / 2;
	if (!guidance)
		skp->Ellipse(iX - iR, iY - iR, iX + iR, iY + iR);


	//two lines for indicating Roll Error
	int iX1 = (int)(1.0 * hps->Markersize * cos(fBankError));
	int iX2 = (int)(InnCoord * hps->Markersize * cos(fBankError));
	int iY1 = (int)(1.0 * hps->Markersize * sin(fBankError));
	int iY2 = (int)(InnCoord * hps->Markersize * sin(fBankError));
	skp->MoveTo(iX + iX1, iY + iY1);
	skp->LineTo(iX + iX2, iY + iY2);

	skp->MoveTo(iX - iX1, iY - iY1);
	skp->LineTo(iX - iX2, iY - iY2);

	//Top Marker for indicating Roll Error
	int iX3 = (int)(0.8 * hps->Markersize * sin(fBankError));
	int iX4 = (int)(InnCoord * hps->Markersize * sin(fBankError));
	int iY3 = (int)(0.8 * hps->Markersize * cos(fBankError));
	int iY4 = (int)(InnCoord * hps->Markersize * cos(fBankError));
	skp->MoveTo(iX + iX3, iY - iY3);
	skp->LineTo(iX + iX4, iY - iY4);

	// draw a diamond, not a circle
	if (guidance) {
		skp->MoveTo(iX + iX4, iY - iY4);
		skp->LineTo(iX + iX2, iY + iY2);
		skp->LineTo(iX - iX4, iY + iY4);
		skp->LineTo(iX - iX2, iY - iY2);
		skp->LineTo(iX + iX4, iY - iY4);
		}
	}

	return true;
}

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new VELCRO (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (VELCRO*)vessel;
}


VELCRO *GetVelcro (HWND hDlg)
{
	// retrieve DG interface from scenario editor
	OBJHANDLE hVessel;
	SendMessage (hDlg, WM_SCNEDITOR, SE_GETVESSEL, (LPARAM)&hVessel);
	return (VELCRO*)oapiGetVesselInterface (hVessel);
}


// --------------------------------------------------------------
// Message procedure for editor page 1 (animation settings)
// --------------------------------------------------------------
BOOL CALLBACK EdPg1Proc (HWND hTab, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
	case WM_COMMAND:
		switch (LOWORD (wParam)) {
		case IDHELP:
			g_hc.topic = "/SE_Anim.htm";
			oapiOpenHelp (&g_hc);
			return TRUE;
/*		case IDC_GEAR_UP:
			GetDG(hTab)->ActivateLandingGear (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_GEAR_DOWN:
			GetDG(hTab)->ActivateLandingGear (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_RETRO_CLOSE:
			GetDG(hTab)->ActivateRCover (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_RETRO_OPEN:
			GetDG(hTab)->ActivateRCover (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_OLOCK_CLOSE:
			GetDG(hTab)->ActivateOuterAirlock (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_OLOCK_OPEN:
			GetDG(hTab)->ActivateOuterAirlock (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_ILOCK_CLOSE:
			GetDG(hTab)->ActivateInnerAirlock (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_ILOCK_OPEN:
			GetDG(hTab)->ActivateInnerAirlock (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_NCONE_CLOSE:
			GetDG(hTab)->ActivateDockingPort (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_NCONE_OPEN:
			GetDG(hTab)->ActivateDockingPort (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_LADDER_RETRACT:
			GetDG(hTab)->ActivateLadder (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_LADDER_EXTEND:
			GetDG(hTab)->ActivateLadder (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_HATCH_CLOSE:
			GetDG(hTab)->ActivateHatch (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_HATCH_OPEN:
			GetDG(hTab)->ActivateHatch (DeltaGlider::DOOR_OPEN);
			return TRUE;
		case IDC_RADIATOR_RETRACT:
			GetDG(hTab)->ActivateRadiator (DeltaGlider::DOOR_CLOSED);
			return TRUE;
		case IDC_RADIATOR_EXTEND:
			GetDG(hTab)->ActivateRadiator (DeltaGlider::DOOR_OPEN);
			return TRUE;
*/		}
		break;
	}
	return FALSE;
}


// --------------------------------------------------------------
// Add vessel-specific pages into scenario editor
// --------------------------------------------------------------
DLLCLBK void secInit (HWND hEditor, OBJHANDLE hVessel)
{
	VELCRO *Velcro = (VELCRO*)oapiGetVesselInterface (hVessel);

	EditorPageSpec eps1 = {"Animations", g_Param.hDLL, IDD_EDITOR_PG1, EdPg1Proc};
	SendMessage (hEditor, WM_SCNEDITOR, SE_ADDPAGEBUTTON, (LPARAM)&eps1);
//	EditorPageSpec eps2 = {"Passengers", g_Param.hDLL, IDD_EDITOR_PG2, EdPg2Proc};
//	SendMessage (hEditor, WM_SCNEDITOR, SE_ADDPAGEBUTTON, (LPARAM)&eps2);
//	if (dg->bDamageEnabled) {
//		EditorPageSpec eps3 = {"Damage", g_Param.hDLL, IDD_EDITOR_PG3, EdPg3Proc};
//		SendMessage (hEditor, WM_SCNEDITOR, SE_ADDPAGEBUTTON, (LPARAM)&eps3);
//	}
}
